"""
LightRay class and member functions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2012 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import h5py
import numpy as np

from yt.funcs import *

from yt.analysis_modules.cosmological_observation.cosmology_splice import \
     CosmologySplice
from yt.analysis_modules.halo_profiler.multi_halo_profiler import \
     HaloProfiler
from yt.convenience import load
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    only_on_root, \
    parallel_objects, \
    parallel_root_only

class LightRay(CosmologySplice):
    """
    Create a LightRay object.  A light ray is much like a light cone,
    in that it stacks together multiple datasets in order to extend a
    redshift interval.  Unlike a light cone, which does randomly
    oriented projections for each dataset, a light ray consists of
    randomly oriented single rays.  The purpose of these is to create
    synthetic QSO lines of sight.

    Once the LightRay object is set up, use LightRay.make_light_ray to
    begin making rays.  Different randomizations can be created with a
    single object by providing different random seeds to make_light_ray.

    Parameters
    ----------
    parameter_filename : string
        The simulation parameter file.
    simulation_type : string
        The simulation type.
    near_redshift : float
        The near (lowest) redshift for the light ray.
    far_redshift : float
        The far (highest) redshift for the light ray.
    use_minimum_datasets : bool
        If True, the minimum number of datasets is used to connect the
        initial and final redshift.  If false, the light ray solution
        will contain as many entries as possible within the redshift
        interval.
        Default: True.
    deltaz_min : float
        Specifies the minimum :math:`\Delta z` between consecutive
        datasets in the returned list.
        Default: 0.0.
    minimum_coherent_box_fraction : float
        Used with use_minimum_datasets set to False, this parameter
        specifies the fraction of the total box size to be traversed
        before rerandomizing the projection axis and center.  This
        was invented to allow light rays with thin slices to sample
        coherent large scale structure, but in practice does not work
        so well.  Try setting this parameter to 1 and see what happens.
        Default: 0.0.
    time_data : bool
        Whether or not to include time outputs when gathering
        datasets for time series.
        Default: True.
    redshift_data : bool
        Whether or not to include redshift outputs when gathering
        datasets for time series.
        Default: True.
    find_outputs : bool
        Whether or not to search for parameter files in the current 
        directory.
        Default: False.

    """
    def __init__(self, parameter_filename, simulation_type,
                 near_redshift, far_redshift,
                 use_minimum_datasets=True, deltaz_min=0.0,
                 minimum_coherent_box_fraction=0.0,
                 time_data=True, redshift_data=True,
                 find_outputs=False):

        self.near_redshift = near_redshift
        self.far_redshift = far_redshift
        self.use_minimum_datasets = use_minimum_datasets
        self.deltaz_min = deltaz_min
        self.minimum_coherent_box_fraction = minimum_coherent_box_fraction

        self.light_ray_solution = []
        self.halo_lists = {}
        self._data = {}

        # Get list of datasets for light ray solution.
        CosmologySplice.__init__(self, parameter_filename, simulation_type,
                                 find_outputs=find_outputs)
        self.light_ray_solution = \
          self.create_cosmology_splice(self.near_redshift, self.far_redshift,
                                       minimal=self.use_minimum_datasets,
                                       deltaz_min=self.deltaz_min,
                                       time_data=time_data,
                                       redshift_data=redshift_data)

    def _calculate_light_ray_solution(self, seed=None, filename=None):
        "Create list of datasets to be added together to make the light ray."

        # Calculate dataset sizes, and get random dataset axes and centers.
        np.random.seed(seed)

        # For box coherence, keep track of effective depth travelled.
        box_fraction_used = 0.0

        for q in range(len(self.light_ray_solution)):
            if (q == len(self.light_ray_solution) - 1):
                z_next = self.near_redshift
            else:
                z_next = self.light_ray_solution[q+1]['redshift']

            # Calculate fraction of box required for a depth of delta z
            self.light_ray_solution[q]['traversal_box_fraction'] = \
                self.cosmology.ComovingRadialDistance(\
                z_next, self.light_ray_solution[q]['redshift']) * \
                self.simulation.hubble_constant / \
                self.simulation.box_size

            # Simple error check to make sure more than 100% of box depth
            # is never required.
            if (self.light_ray_solution[q]['traversal_box_fraction'] > 1.0):
                mylog.error("Warning: box fraction required to go from z = %f to %f is %f" %
                            (self.light_ray_solution[q]['redshift'], z_next,
                             self.light_ray_solution[q]['traversal_box_fraction']))
                mylog.error("Full box delta z is %f, but it is %f to the next data dump." %
                            (self.light_ray_solution[q]['deltazMax'],
                             self.light_ray_solution[q]['redshift']-z_next))

            # Get dataset axis and center.
            # If using box coherence, only get start point and vector if
            # enough of the box has been used,
            # or if box_fraction_used will be greater than 1 after this slice.
            if (q == 0) or (self.minimum_coherent_box_fraction == 0) or \
                    (box_fraction_used >
                     self.minimum_coherent_box_fraction) or \
                    (box_fraction_used +
                     self.light_ray_solution[q]['traversal_box_fraction'] > 1.0):
                # Random start point
                self.light_ray_solution[q]['start'] = np.random.random(3)
                theta = np.pi * np.random.random()
                phi = 2 * np.pi * np.random.random()
                box_fraction_used = 0.0
            else:
                # Use end point of previous segment and same theta and phi.
                self.light_ray_solution[q]['start'] = \
                  self.light_ray_solution[q-1]['end'][:]

            self.light_ray_solution[q]['end'] = \
              self.light_ray_solution[q]['start'] + \
                self.light_ray_solution[q]['traversal_box_fraction'] * \
                np.array([np.cos(phi) * np.sin(theta),
                          np.sin(phi) * np.sin(theta),
                          np.cos(theta)])
            box_fraction_used += \
              self.light_ray_solution[q]['traversal_box_fraction']

        if filename is not None:
            self._write_light_ray_solution(filename,
                extra_info={'parameter_filename':self.parameter_filename,
                            'random_seed':seed,
                            'far_redshift':self.far_redshift,
                            'near_redshift':self.near_redshift})

    def make_light_ray(self, seed=None, fields=None,
                       solution_filename=None, data_filename=None,
                       get_los_velocity=False,
                       get_nearest_halo=False,
                       nearest_halo_fields=None,
                       halo_list_file=None,
                       halo_profiler_parameters=None,
                       njobs=1, dynamic=False):
        """
        Create a light ray and get field values for each lixel.  A light
        ray consists of a list of field values for cells intersected by
        the ray and the path length of the ray through those cells.
        Light ray data can be written out to an hdf5 file.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
            Default: None.
        fields : list
            A list of fields for which to get data.
            Default: None.
        solution_filename : string
            Path to a text file where the trajectories of each
            subray is written out.
            Default: None.
        data_filename : string
            Path to output file for ray data.
            Default: None.
        get_los_velocity : bool
            If True, the line of sight velocity is calculated for
            each point in the ray.
            Default: False.
        get_nearest_halo : bool
            If True, the HaloProfiler will be used to calculate the
            distance and mass of the nearest halo for each point in the
            ray.  This option requires additional information to be
            included.  See below for an example.
            Default: False.
        nearest_halo_fields : list
            A list of fields to be calculated for the halos nearest to
            every lixel in the ray.
            Default: None.
        halo_list_file : str
            Filename containing a list of halo properties to be used 
            for getting the nearest halos to absorbers.
            Default: None.
        halo_profiler_parameters: dict
            A dictionary of parameters to be passed to the HaloProfiler
            to create the appropriate data used to get properties for
            the nearest halos.
            Default: None.
        njobs : int
            The number of parallel jobs over which the slices for the
            halo mask will be split.  Choose -1 for one processor per
            individual slice and 1 to have all processors work together
            on each projection.
            Default: 1
        dynamic : bool
            If True, use dynamic load balancing to create the projections.
            Default: False.

        Notes
        -----

        The light ray tool will use the HaloProfiler to calculate the
        distance and mass of the nearest halo to that pixel.  In order
        to do this, a dictionary called halo_profiler_parameters is used
        to pass instructions to the HaloProfiler.  This dictionary has
        three additional keywords.

        halo_profiler_kwargs : dict
            A dictionary of standard HaloProfiler keyword
            arguments and values to be given to the HaloProfiler.

        halo_profiler_actions : list
            A list of actions to be performed by the HaloProfiler.
            Each item in the list should be a dictionary with the following
            entries: "function", "args", and "kwargs", for the function to
            be performed, the arguments supplied to that function, and the
            keyword arguments.

        halo_list : string
            'all' to use the full halo list, or 'filtered' to use the
            filtered halo list created after calling make_profiles.

        Examples
        --------

        >>> from yt.mods import *
        >>> from yt.analysis_modules.halo_profiler.api import *
        >>> from yt.analysis_modules.cosmological_analysis.light_ray.api import LightRay
        >>> halo_profiler_kwargs = {'halo_list_file': 'HopAnalysis.out'}
        >>> halo_profiler_actions = []
        >>> # Add a virial filter.
        >>> halo_profiler_actions.append({'function': add_halo_filter,
        ...                           'args': VirialFilter,
        ...                           'kwargs': {'overdensity_field': 'ActualOverdensity',
        ...                                      'virial_overdensity': 200,
        ...                                      'virial_filters': [['TotalMassMsun','>=','1e14']],
        ...                                      'virial_quantities': ['TotalMassMsun','RadiusMpc']}})
        ...
        >>> # Make the profiles.
        >>> halo_profiler_actions.append({'function': make_profiles,
        ...                           'args': None,
        ...                           'kwargs': {'filename': 'VirializedHalos.h5'}})
        ...
        >>> halo_list = 'filtered'
        >>> halo_profiler_parameters = dict(halo_profiler_kwargs=halo_profiler_kwargs,
        ...                             halo_profiler_actions=halo_profiler_actions,
        ...                             halo_list=halo_list)
        ...
        >>> my_ray = LightRay('simulation.par', 'Enzo', 0., 0.1,
        ...                use_minimum_datasets=True,
        ...                time_data=False)
        ...
        >>> my_ray.make_light_ray(seed=12345,
        ...                   solution_filename='solution.txt',
        ...                   data_filename='my_ray.h5',
        ...                   fields=['Temperature', 'Density'],
        ...                   get_nearest_halo=True,
        ...                   nearest_halo_fields=['TotalMassMsun_100',
        ...                                        'RadiusMpc_100'],
        ...                   halo_list_file='VirializedHalos.h5',
        ...                   halo_profiler_parameters=halo_profiler_parameters,
        ...                   get_los_velocity=True)
        
        """

        if halo_profiler_parameters is None:
            halo_profiler_parameters = {}
        if nearest_halo_fields is None:
            nearest_halo_fields = []

        # Calculate solution.
        self._calculate_light_ray_solution(seed=seed, filename=solution_filename)

        # Initialize data structures.
        self._data = {}
        if fields is None: fields = []
        data_fields = fields[:]
        all_fields = fields[:]
        all_fields.extend(['dl', 'dredshift', 'redshift'])
        if get_nearest_halo:
            all_fields.extend(['x', 'y', 'z', 'nearest_halo'])
            all_fields.extend(['nearest_halo_%s' % field \
                               for field in nearest_halo_fields])
            data_fields.extend(['x', 'y', 'z'])
        if get_los_velocity:
            all_fields.extend(['x-velocity', 'y-velocity',
                               'z-velocity', 'los_velocity'])
            data_fields.extend(['x-velocity', 'y-velocity', 'z-velocity'])

        all_ray_storage = {}
        for my_storage, my_segment in parallel_objects(self.light_ray_solution,
                                                       storage=all_ray_storage,
                                                       njobs=njobs, dynamic=dynamic):
            mylog.info("Creating ray segment at z = %f." %
                       my_segment['redshift'])
            if my_segment['next'] is None:
                next_redshift = self.near_redshift
            else:
                next_redshift = my_segment['next']['redshift']

            mylog.info("Getting segment at z = %s: %s to %s." %
                       (my_segment['redshift'], my_segment['start'],
                        my_segment['end']))

            # Load dataset for segment.
            pf = load(my_segment['filename'])

            # Break periodic ray into non-periodic segments.
            sub_segments = periodic_ray(my_segment['start'], my_segment['end'])

            # Prepare data structure for subsegment.
            sub_data = {}
            sub_data['segment_redshift'] = my_segment['redshift']
            for field in all_fields:
                sub_data[field] = np.array([])

            # Get data for all subsegments in segment.
            for sub_segment in sub_segments:
                mylog.info("Getting subsegment: %s to %s." %
                           (list(sub_segment[0]), list(sub_segment[1])))
                sub_ray = pf.h.ray(sub_segment[0], sub_segment[1])
                sub_data['dl'] = np.concatenate([sub_data['dl'],
                                                 (sub_ray['dts'] *
                                                  vector_length(sub_segment[0],
                                                                sub_segment[1]))])
                for field in data_fields:
                    sub_data[field] = np.concatenate([sub_data[field],
                                                      (sub_ray[field])])

                if get_los_velocity:
                    line_of_sight = sub_segment[1] - sub_segment[0]
                    line_of_sight /= ((line_of_sight**2).sum())**0.5
                    sub_vel = np.array([sub_ray['x-velocity'],
                                        sub_ray['y-velocity'],
                                        sub_ray['z-velocity']])
                    sub_data['los_velocity'] = \
                      np.concatenate([sub_data['los_velocity'],
                                      (np.rollaxis(sub_vel, 1) *
                                       line_of_sight).sum(axis=1)])
                    del sub_vel

                sub_ray.clear_data()
                del sub_ray

            # Get redshift for each lixel.  Assume linear relation between l and z.
            sub_data['dredshift'] = (my_segment['redshift'] - next_redshift) * \
                (sub_data['dl'] / vector_length(my_segment['start'], my_segment['end']))
            sub_data['redshift'] = my_segment['redshift'] \
                - sub_data['dredshift'].cumsum() + sub_data['dredshift']

            # Calculate distance to nearest object on halo list for each lixel.
            if get_nearest_halo:
                halo_list = self._get_halo_list(pf, fields=nearest_halo_fields,
                                                filename=halo_list_file,
                                                **halo_profiler_parameters)
                sub_data.update(self._get_nearest_halo_properties(sub_data, halo_list,
                                fields=nearest_halo_fields))
                sub_data['nearest_halo'] *= pf.units['mpccm']

            # Remove empty lixels.
            sub_dl_nonzero = sub_data['dl'].nonzero()
            for field in all_fields:
                sub_data[field] = sub_data[field][sub_dl_nonzero]
            del sub_dl_nonzero

            # Convert dl to cm.
            sub_data['dl'] *= pf.units['cm']

            # Add to storage.
            my_storage.result = sub_data

            pf.h.clear_all_data()
            del pf

        # Reconstruct ray data from parallel_objects storage.
        all_data = [my_data for my_data in all_ray_storage.values()]
        # This is now a list of segments where each one is a dictionary
        # with all the fields.
        all_data.sort(key=lambda a:a['segment_redshift'], reverse=True)
        # Flatten the list into a single dictionary containing fields
        # for the whole ray.
        all_data = _flatten_dict_list(all_data, exceptions=['segment_redshift'])

        if data_filename is not None:
            self._write_light_ray(data_filename, all_data)

        self._data = all_data
        return all_data

    def _get_halo_list(self, pf, fields=None, filename=None, 
                       halo_profiler_kwargs=None,
                       halo_profiler_actions=None, halo_list='all'):
        "Load a list of halos for the pf."

        if str(pf) in self.halo_lists:
            return self.halo_lists[str(pf)]

        if fields is None: fields = []

        if filename is not None and \
                os.path.exists(os.path.join(pf.fullpath, filename)):

            my_filename = os.path.join(pf.fullpath, filename)
            mylog.info("Loading halo list from %s." % my_filename)
            my_list = {}
            in_file = h5py.File(my_filename, 'r')
            for field in fields + ['center']:
                my_list[field] = in_file[field][:]
            in_file.close()

        else:
            my_list = self._halo_profiler_list(pf, fields=fields,
                                               halo_profiler_kwargs=halo_profiler_kwargs,
                                               halo_profiler_actions=halo_profiler_actions,
                                               halo_list=halo_list)

        self.halo_lists[str(pf)] = my_list
        return self.halo_lists[str(pf)]

    def _halo_profiler_list(self, pf, fields=None, 
                            halo_profiler_kwargs=None,
                            halo_profiler_actions=None, halo_list='all'):
        "Run the HaloProfiler to get the halo list."

        if halo_profiler_kwargs is None: halo_profiler_kwargs = {}
        if halo_profiler_actions is None: halo_profiler_actions = []

        hp = HaloProfiler(pf, **halo_profiler_kwargs)
        for action in halo_profiler_actions:
            if not action.has_key('args'): action['args'] = ()
            if not action.has_key('kwargs'): action['kwargs'] = {}
            action['function'](hp, *action['args'], **action['kwargs'])

        if halo_list == 'all':
            hp_list = copy.deepcopy(hp.all_halos)
        elif halo_list == 'filtered':
            hp_list = copy.deepcopy(hp.filtered_halos)
        else:
            mylog.error("Keyword, halo_list, must be either 'all' or 'filtered'.")
            hp_list = None

        del hp

        # Create position array from halo list.
        return_list = dict([(field, []) for field in fields + ['center']])
        for halo in hp_list:
            for field in fields + ['center']:
                return_list[field].append(halo[field])
        for field in fields + ['center']:
            return_list[field] = np.array(return_list[field])
        return return_list
        
    def _get_nearest_halo_properties(self, data, halo_list, fields=None):
        """
        Calculate distance to nearest object in halo list for each lixel in data.
        Return list of distances and other properties of nearest objects.
        """

        if fields is None: fields = []
        field_data = dict([(field, np.zeros_like(data['x'])) \
                           for field in fields])
        nearest_distance = np.zeros_like(data['x'])

        if halo_list['center'].size > 0:
            for index in xrange(nearest_distance.size):
                nearest = np.argmin(periodic_distance(np.array([data['x'][index],
                                                                data['y'][index],
                                                                data['z'][index]]),
                                                      halo_list['center']))
                nearest_distance[index] = periodic_distance(np.array([data['x'][index],
                                                                      data['y'][index],
                                                                      data['z'][index]]),
                                                            halo_list['center'][nearest])
                for field in fields:
                    field_data[field][index] = halo_list[field][nearest]

        return_data = {'nearest_halo': nearest_distance}
        for field in fields:
            return_data['nearest_halo_%s' % field] = field_data[field]
        return return_data

    @parallel_root_only
    def _write_light_ray(self, filename, data):
        "Write light ray data to hdf5 file."

        mylog.info("Saving light ray data to %s." % filename)
        output = h5py.File(filename, 'w')
        for field in data.keys():
            output.create_dataset(field, data=data[field])
        output.close()

    @parallel_root_only
    def _write_light_ray_solution(self, filename, extra_info=None):
        "Write light ray solution to a file."

        mylog.info("Writing light ray solution to %s." % filename)
        f = open(filename, 'w')
        if extra_info is not None:
            for par, val in extra_info.items():
                f.write("%s = %s\n" % (par, val))
        f.write("\nSegment Redshift dl/box    Start x       y             z             End x         y             z            Dataset\n")
        for q, my_segment in enumerate(self.light_ray_solution):
            f.write("%04d    %.6f %.6f % .10f % .10f % .10f % .10f % .10f % .10f %s\n" % \
                        (q, my_segment['redshift'], my_segment['traversal_box_fraction'],
                         my_segment['start'][0], my_segment['start'][1], my_segment['start'][2],
                         my_segment['end'][0], my_segment['end'][1], my_segment['end'][2],
                         my_segment['filename']))
        f.close()

def _flatten_dict_list(data, exceptions=None):
    "Flatten the list of dicts into one dict."

    if exceptions is None: exceptions = []
    new_data = {}
    for datum in data:
        for field in [field for field in datum.keys()
                      if field not in exceptions]:
            if field in new_data:
                new_data[field] = np.concatenate([new_data[field], datum[field]])
            else:
                new_data[field] = np.copy(datum[field])
    return new_data

def vector_length(start, end):
    "Calculate vector length."

    return np.sqrt(np.power((end - start), 2).sum())

def periodic_distance(coord1, coord2):
    "Calculate length of shortest vector between to points in periodic domain."
    dif = coord1 - coord2

    dim = np.ones(coord1.shape,dtype=int)
    def periodic_bind(num):
        pos = np.abs(num % dim)
        neg = np.abs(num % -dim)
        return np.min([pos,neg],axis=0)

    dif = periodic_bind(dif)
    return np.sqrt((dif * dif).sum(axis=-1))

def periodic_ray(start, end, left=None, right=None):
    "Break up periodic ray into non-periodic segments."

    if left is None:
        left = np.zeros(start.shape)
    if right is None:
        right = np.ones(start.shape)
    dim = right - left

    vector = end - start
    wall = np.zeros(start.shape)
    close = np.zeros(start.shape, dtype=object)

    left_bound = vector < 0
    right_bound = vector > 0
    no_bound = vector == 0.0
    bound = vector != 0.0

    wall[left_bound] = left[left_bound]
    close[left_bound] = np.max
    wall[right_bound] = right[right_bound]
    close[right_bound] = np.min
    wall[no_bound] = np.inf
    close[no_bound] = np.min

    segments = []
    this_start = np.copy(start)
    this_end = np.copy(end)
    t = 0.0
    tolerance = 1e-6

    while t < 1.0 - tolerance:
        hit_left = (this_start <= left) & (vector < 0)
        if (hit_left).any():
            this_start[hit_left] += dim[hit_left]
            this_end[hit_left] += dim[hit_left]
        hit_right = (this_start >= right) & (vector > 0)
        if (hit_right).any():
            this_start[hit_right] -= dim[hit_right]
            this_end[hit_right] -= dim[hit_right]

        nearest = np.array([close[q]([this_end[q], wall[q]]) \
                                for q in range(start.size)])
        dt = ((nearest - this_start) / vector)[bound].min()
        now = this_start + vector * dt
        close_enough = np.abs(now - nearest) < 1e-10
        now[close_enough] = nearest[close_enough]
        segments.append([np.copy(this_start), np.copy(now)])
        this_start = np.copy(now)
        t += dt

    return segments



"""
Data structures for Enzo

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import weakref
import numpy as np
import os
import stat
import string
import re
try:
    from pyhdf_np import SD
except ImportError:
    pass

from itertools import izip

from yt.funcs import *
from yt.config import ytcfg
from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.data_objects.hierarchy import \
    AMRHierarchy
from yt.data_objects.static_output import \
    StaticOutput
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion
from yt.utilities import hdf5_light_reader
from yt.utilities.logger import ytLogger as mylog

from .definitions import parameterDict
from .fields import \
    EnzoFieldInfo, Enzo2DFieldInfo, Enzo1DFieldInfo, \
    add_enzo_field, add_enzo_2d_field, add_enzo_1d_field, \
    KnownEnzoFields

from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_blocking_call

class EnzoGrid(AMRGridPatch):
    """
    Class representing a single Enzo Grid instance.
    """

    __slots__ = []
    def __init__(self, id, hierarchy):
        """
        Returns an instance of EnzoGrid with *id*, associated with
        *filename* and *hierarchy*.
        """
        #All of the field parameters will be passed to us as needed.
        AMRGridPatch.__init__(self, id, filename = None, hierarchy = hierarchy)
        self._children_ids = []
        self._parent_id = -1
        self.Level = -1

    def _guess_properties_from_parent(self):
        """
        We know that our grid boundary occurs on the cell boundary of our
        parent.  This can be a very expensive process, but it is necessary
        in some hierarchys, where yt is unable to generate a completely
        space-filling tiling of grids, possibly due to the finite accuracy in a
        standard Enzo hierarchy file.
        """
        rf = self.pf.refine_by
        my_ind = self.id - self._id_offset
        le = self.LeftEdge
        self.dds = self.Parent.dds/rf
        ParentLeftIndex = np.rint((self.LeftEdge-self.Parent.LeftEdge)/self.Parent.dds)
        self.start_index = rf*(ParentLeftIndex + self.Parent.get_global_startindex()).astype('int64')
        self.LeftEdge = self.Parent.LeftEdge + self.Parent.dds * ParentLeftIndex
        self.RightEdge = self.LeftEdge + self.ActiveDimensions*self.dds
        self.hierarchy.grid_left_edge[my_ind,:] = self.LeftEdge
        self.hierarchy.grid_right_edge[my_ind,:] = self.RightEdge
        self._child_mask = None
        self._child_index_mask = None
        self._child_indices = None
        self._setup_dx()

    def set_filename(self, filename):
        """
        Intelligently set the filename.
        """
        if filename is None:
            self.filename = filename
            return
        if self.hierarchy._strip_path:
            self.filename = os.path.join(self.hierarchy.directory,
                                         os.path.basename(filename))
        elif filename[0] == os.path.sep:
            self.filename = filename
        else:
            self.filename = os.path.join(self.hierarchy.directory, filename)
        return

    def __repr__(self):
        return "EnzoGrid_%04i" % (self.id)

    @property
    def Parent(self):
        if self._parent_id == -1: return None
        return self.hierarchy.grids[self._parent_id - self._id_offset]

    @property
    def Children(self):
        return [self.hierarchy.grids[cid - self._id_offset]
                for cid in self._children_ids]

class EnzoGridInMemory(EnzoGrid):
    __slots__ = ['proc_num']
    def set_filename(self, filename):
        pass

class EnzoGridGZ(EnzoGrid):

    __slots__ = ()

    def retrieve_ghost_zones(self, n_zones, fields, all_levels=False,
                             smoothed=False):
        # We ignore smoothed in this case.
        if n_zones > 3:
            return EnzoGrid.retrieve_ghost_zones(
                self, n_zones, fields, all_levels, smoothed)
        # ----- Below is mostly the original code, except we remove the field
        # ----- access section
        # We will attempt this by creating a datacube that is exactly bigger
        # than the grid by nZones*dx in each direction
        nl = self.get_global_startindex() - n_zones
        nr = nl + self.ActiveDimensions + 2*n_zones
        new_left_edge = nl * self.dds + self.pf.domain_left_edge
        new_right_edge = nr * self.dds + self.pf.domain_left_edge
        # Something different needs to be done for the root grid, though
        level = self.Level
        args = (level, new_left_edge, new_right_edge)
        kwargs = {'dims': self.ActiveDimensions + 2*n_zones,
                  'num_ghost_zones':n_zones,
                  'use_pbar':False}
        # This should update the arguments to set the field parameters to be
        # those of this grid.
        kwargs.update(self.field_parameters)
        if smoothed:
            #cube = self.hierarchy.smoothed_covering_grid(
            #    level, new_left_edge, new_right_edge, **kwargs)
            cube = self.hierarchy.smoothed_covering_grid(
                level, new_left_edge, **kwargs)
        else:
            cube = self.hierarchy.covering_grid(
                level, new_left_edge, **kwargs)
        # ----- This is EnzoGrid.get_data, duplicated here mostly for
        # ----  efficiency's sake.
        sl = [slice(3 - n_zones, -(3 - n_zones)) for i in range(3)]
        if fields is None: return cube
        for field in ensure_list(fields):
            if field in self.hierarchy.field_list:
                conv_factor = 1.0
                if self.pf.field_info.has_key(field):
                    conv_factor = self.pf.field_info[field]._convert_function(self)
                if self.pf.field_info[field].particle_type: continue
                temp = self.hierarchy.io._read_raw_data_set(self, field)
                temp = temp.swapaxes(0, 2)
                cube.field_data[field] = np.multiply(temp, conv_factor, temp)[sl]
        return cube

class EnzoHierarchy(AMRHierarchy):

    _strip_path = False
    grid = EnzoGrid

    def __init__(self, pf, data_style):
        
        self.data_style = data_style
        if pf.file_style != None:
            self._bn = pf.file_style
        else:
            self._bn = "%s.cpu%%04i"
        self.hierarchy_filename = os.path.abspath(
            "%s.hierarchy" % (pf.parameter_filename))
        harray_fn = self.hierarchy_filename[:-9] + "harrays"
        if ytcfg.getboolean("yt","serialize") and os.path.exists(harray_fn):
            try:
                harray_fp = h5py.File(harray_fn)
                self.num_grids = harray_fp["/Level"].len()
                harray_fp.close()
            except IOError:
                pass
        elif os.path.getsize(self.hierarchy_filename) == 0:
            raise IOError(-1,"File empty", self.hierarchy_filename)
        self.directory = os.path.dirname(self.hierarchy_filename)

        # For some reason, r8 seems to want Float64
        if pf.has_key("CompilerPrecision") \
            and pf["CompilerPrecision"] == "r4":
            self.float_type = 'float32'
        else:
            self.float_type = 'float64'

        AMRHierarchy.__init__(self, pf, data_style)
        # sync it back
        self.parameter_file.data_style = self.data_style

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        test_grid = test_grid_id = None
        self.num_stars = 0
        for line in rlines(open(self.hierarchy_filename, "rb")):
            if line.startswith("BaryonFileName") or \
               line.startswith("ParticleFileName") or \
               line.startswith("FileName "):
                test_grid = line.split("=")[-1].strip().rstrip()
            if line.startswith("NumberOfStarParticles"):
                self.num_stars = int(line.split("=")[-1])
            if line.startswith("Grid "):
                self.num_grids = test_grid_id = int(line.split("=")[-1])
                break
        self._guess_data_style(self.pf.dimensionality, test_grid, test_grid_id)

    def _guess_data_style(self, rank, test_grid, test_grid_id):
        if test_grid[0] != os.path.sep:
            test_grid = os.path.join(self.directory, test_grid)
        if not os.path.exists(test_grid):
            test_grid = os.path.join(self.directory,
                                    os.path.basename(test_grid))
            mylog.debug("Your data uses the annoying hardcoded path.")
            self._strip_path = True
        if self.data_style is not None: return
        try:
            a = SD.SD(test_grid)
            self.data_style = 'enzo_hdf4'
            mylog.debug("Detected HDF4")
        except:
            try:
                list_of_sets = hdf5_light_reader.ReadListOfDatasets(test_grid, "/")
            except:
                print "Could not find dataset.  Defaulting to packed HDF5"
                list_of_sets = []
            if len(list_of_sets) == 0 and rank == 3:
                mylog.debug("Detected packed HDF5")
                if self.parameters.get("WriteGhostZones", 0) == 1:
                    self.data_style= "enzo_packed_3d_gz"
                    self.grid = EnzoGridGZ
                else:
                    self.data_style = 'enzo_packed_3d'
            elif len(list_of_sets) > 0 and rank == 3:
                mylog.debug("Detected unpacked HDF5")
                self.data_style = 'enzo_hdf5'
            elif len(list_of_sets) == 0 and rank == 2:
                mylog.debug("Detect packed 2D")
                self.data_style = 'enzo_packed_2d'
            elif len(list_of_sets) == 0 and rank == 1:
                mylog.debug("Detect packed 1D")
                self.data_style = 'enzo_packed_1d'
            else:
                raise TypeError

    # Sets are sorted, so that won't work!
    def _parse_hierarchy(self):
        def _next_token_line(token, f):
            for line in f:
                if line.startswith(token):
                    return line.split()[2:]
        if os.path.exists(self.hierarchy_filename[:-9] + "harrays"):
            if self._parse_binary_hierarchy(): return
        t1 = time.time()
        pattern = r"Pointer: Grid\[(\d*)\]->NextGrid(Next|This)Level = (\d*)\s+$"
        patt = re.compile(pattern)
        f = open(self.hierarchy_filename, "rb")
        self.grids = [self.grid(1, self)]
        self.grids[0].Level = 0
        si, ei, LE, RE, fn, npart = [], [], [], [], [], []
        all = [si, ei, LE, RE, fn]
        pbar = get_pbar("Parsing Hierarchy", self.num_grids)
        for grid_id in xrange(self.num_grids):
            pbar.update(grid_id)
            # We will unroll this list
            si.append(_next_token_line("GridStartIndex", f))
            ei.append(_next_token_line("GridEndIndex", f))
            LE.append(_next_token_line("GridLeftEdge", f))
            RE.append(_next_token_line("GridRightEdge", f))
            nb = int(_next_token_line("NumberOfBaryonFields", f)[0])
            fn.append([None])
            if nb > 0: fn[-1] = _next_token_line("BaryonFileName", f)
            npart.append(int(_next_token_line("NumberOfParticles", f)[0]))
            if nb == 0 and npart[-1] > 0: fn[-1] = _next_token_line("ParticleFileName", f)
            for line in f:
                if len(line) < 2: break
                if line.startswith("Pointer:"):
                    vv = patt.findall(line)[0]
                    self.__pointer_handler(vv)
        pbar.finish()
        self._fill_arrays(ei, si, LE, RE, npart)
        temp_grids = np.empty(self.num_grids, dtype='object')
        temp_grids[:] = self.grids
        self.grids = temp_grids
        self.filenames = fn
        self._store_binary_hierarchy()
        t2 = time.time()

    def _fill_arrays(self, ei, si, LE, RE, npart):
        self.grid_dimensions.flat[:] = ei
        self.grid_dimensions -= np.array(si, self.float_type)
        self.grid_dimensions += 1
        self.grid_left_edge.flat[:] = LE
        self.grid_right_edge.flat[:] = RE
        self.grid_particle_count.flat[:] = npart

    def __pointer_handler(self, m):
        sgi = int(m[2])-1
        if sgi == -1: return # if it's 0, then we're done with that lineage
        # Okay, so, we have a pointer.  We make a new grid, with an id of the length+1
        # (recall, Enzo grids are 1-indexed)
        self.grids.append(self.grid(len(self.grids)+1, self))
        # We'll just go ahead and make a weakref to cache
        second_grid = self.grids[sgi] # zero-indexed already
        first_grid = self.grids[int(m[0])-1]
        if m[1] == "Next":
            first_grid._children_ids.append(second_grid.id)
            second_grid._parent_id = first_grid.id
            second_grid.Level = first_grid.Level + 1
        elif m[1] == "This":
            if first_grid.Parent is not None:
                first_grid.Parent._children_ids.append(second_grid.id)
                second_grid._parent_id = first_grid._parent_id
            second_grid.Level = first_grid.Level
        self.grid_levels[sgi] = second_grid.Level

    def _parse_binary_hierarchy(self):
        mylog.info("Getting the binary hierarchy")
        if not ytcfg.getboolean("yt","serialize"): return False
        try:
            f = h5py.File(self.hierarchy_filename[:-9] + "harrays")
        except:
            return False
        hash = f["/"].attrs.get("hash", None)
        if hash != self.parameter_file._hash():
            mylog.info("Binary hierarchy does not match: recreating")
            f.close()
            return False
        self.grid_dimensions[:] = f["/ActiveDimensions"][:]
        self.grid_left_edge[:] = f["/LeftEdges"][:]
        self.grid_right_edge[:] = f["/RightEdges"][:]
        self.grid_particle_count[:,0] = f["/NumberOfParticles"][:]
        levels = f["/Level"][:]
        parents = f["/ParentIDs"][:]
        procs = f["/Processor"][:]
        grids = []
        self.filenames = []
        grids = [self.grid(gi+1, self) for gi in xrange(self.num_grids)]
        giter = izip(grids, levels, procs, parents)
        bn = self._bn % (self.pf)
        pmap = [(bn % P,) for P in xrange(procs.max()+1)]
        pmap.append((None, )) # Now, P==-1 will give None
        for grid,L,P,Pid in giter:
            grid.Level = L
            grid._parent_id = Pid
            if Pid > -1:
                grids[Pid-1]._children_ids.append(grid.id)
            self.filenames.append(pmap[P])
        self.grids = np.array(grids, dtype='object')
        f.close()
        mylog.info("Finished with binary hierarchy reading")
        return True

    @parallel_blocking_call
    def _store_binary_hierarchy(self):
        # We don't do any of the logic here, we just check if the data file
        # is open...
        if self._data_file is None: return
        if self._data_mode == 'r': return
        if self.data_style != "enzo_packed_3d": return
        mylog.info("Storing the binary hierarchy")
        try:
            f = h5py.File(self.hierarchy_filename[:-9] + "harrays", "w")
        except IOError:
            return
        f["/"].attrs["hash"] = self.parameter_file._hash()
        f.create_dataset("/LeftEdges", data=self.grid_left_edge)
        f.create_dataset("/RightEdges", data=self.grid_right_edge)
        parents, procs, levels = [], [], []
        for i,g in enumerate(self.grids):
            if g.Parent is not None:
                parents.append(g.Parent.id)
            else:
                parents.append(-1)
            if self.filenames[i][0] is None:
                procs.append(-1)
            else:
                procs.append(int(self.filenames[i][0][-4:]))
            levels.append(g.Level)

        parents = np.array(parents, dtype='int64')
        procs = np.array(procs, dtype='int64')
        levels = np.array(levels, dtype='int64')
        f.create_dataset("/ParentIDs", data=parents)
        f.create_dataset("/Processor", data=procs)
        f.create_dataset("/Level", data=levels)

        f.create_dataset("/ActiveDimensions", data=self.grid_dimensions)
        f.create_dataset("/NumberOfParticles", data=self.grid_particle_count[:,0])

        f.close()

    def _rebuild_top_grids(self, level = 0):
        #for level in xrange(self.max_level+1):
        mylog.info("Rebuilding grids on level %s", level)
        cmask = (self.grid_levels.flat == (level + 1))
        cmsum = cmask.sum()
        mask = np.zeros(self.num_grids, dtype='bool')
        for grid in self.select_grids(level):
            mask[:] = 0
            LE = self.grid_left_edge[grid.id - grid._id_offset]
            RE = self.grid_right_edge[grid.id - grid._id_offset]
            grids, grid_i = self.get_box_grids(LE, RE)
            mask[grid_i] = 1
            grid._children_ids = []
            cgrids = self.grids[ ( mask * cmask).astype('bool') ]
            mylog.info("%s: %s / %s", grid, len(cgrids), cmsum)
            for cgrid in cgrids:
                grid._children_ids.append(cgrid.id)
                cgrid._parent_id = grid.id
        mylog.info("Finished rebuilding")

    def _populate_grid_objects(self):
        reconstruct = ytcfg.getboolean("yt","reconstruct_hierarchy")
        for g,f in izip(self.grids, self.filenames):
            g._prepare_grid()
            g._setup_dx()
            g.set_filename(f[0])
            if reconstruct:
                if g.Parent is not None: g._guess_properties_from_parent()
        del self.filenames # No longer needed.
        self.max_level = self.grid_levels.max()

    def _detect_fields(self):
        self.field_list = []
        # Do this only on the root processor to save disk work.
        if self.comm.rank == 0 or self.comm.rank == None:
            field_list = self.get_data("/", "DataFields")
            if field_list is None:
                mylog.info("Gathering a field list (this may take a moment.)")
                field_list = set()
                random_sample = self._generate_random_grids()
                for grid in random_sample:
                    if not hasattr(grid, 'filename'): continue
                    try:
                        gf = self.io._read_field_names(grid)
                    except self.io._read_exception:
                        mylog.debug("Grid %s is a bit funky?", grid.id)
                        continue
                    mylog.debug("Grid %s has: %s", grid.id, gf)
                    field_list = field_list.union(gf)
        else:
            field_list = None
        field_list = self.comm.mpi_bcast(field_list)
        self.save_data(list(field_list),"/","DataFields",passthrough=True)
        self.field_list = list(field_list)

    def _generate_random_grids(self):
        if self.num_grids > 40:
            starter = np.random.randint(0, 20)
            random_sample = np.mgrid[starter:len(self.grids)-1:20j].astype("int32")
            # We also add in a bit to make sure that some of the grids have
            # particles
            gwp = self.grid_particle_count > 0
            if np.any(gwp) and not np.any(gwp[(random_sample,)]):
                # We just add one grid.  This is not terribly efficient.
                first_grid = np.where(gwp)[0][0]
                random_sample.resize((21,))
                random_sample[-1] = first_grid
                mylog.debug("Added additional grid %s", first_grid)
            mylog.debug("Checking grids: %s", random_sample.tolist())
        else:
            random_sample = np.mgrid[0:max(len(self.grids),1)].astype("int32")
        return self.grids[(random_sample,)]

    def find_particles_by_type(self, ptype, max_num=None, additional_fields=None):
        """
        Returns a structure of arrays with all of the particles'
        positions, velocities, masses, types, IDs, and attributes for
        a particle type **ptype** for a maximum of **max_num**
        particles.  If non-default particle fields are used, provide
        them in **additional_fields**.
        """
        # Not sure whether this routine should be in the general HierarchyType.
        if self.grid_particle_count.sum() == 0:
            mylog.info("Data contains no particles.");
            return None
        if additional_fields is None:
            additional_fields = ['metallicity_fraction', 'creation_time',
                                 'dynamical_time']
        pfields = [f for f in self.field_list if f.startswith('particle_')]
        nattr = self.parameter_file['NumberOfParticleAttributes']
        if nattr > 0:
            pfields += additional_fields[:nattr]
        # Find where the particles reside and count them
        if max_num is None: max_num = 1e100
        total = 0
        pstore = []
        for level in range(self.max_level, -1, -1):
            for grid in self.select_grids(level):
                index = np.where(grid['particle_type'] == ptype)[0]
                total += len(index)
                pstore.append(index)
                if total >= max_num: break
            if total >= max_num: break
        result = None
        if total > 0:
            result = {}
            for p in pfields:
                result[p] = np.zeros(total, 'float64')
            # Now we retrieve data for each field
            ig = count = 0
            for level in range(self.max_level, -1, -1):
                for grid in self.select_grids(level):
                    nidx = len(pstore[ig])
                    if nidx > 0:
                        for p in pfields:
                            result[p][count:count+nidx] = grid[p][pstore[ig]]
                        count += nidx
                    ig += 1
                    if count >= total: break
                if count >= total: break
            # Crop data if retrieved more than max_num
            if count > max_num:
                for p in pfields:
                    result[p] = result[p][0:max_num]
        return result


class EnzoHierarchyInMemory(EnzoHierarchy):

    grid = EnzoGridInMemory
    _enzo = None

    @property
    def enzo(self):
        if self._enzo is None:
            import enzo
            self._enzo = enzo
        return self._enzo

    def __init__(self, pf, data_style = None):
        self.data_style = data_style
        self.float_type = 'float64'
        self.parameter_file = weakref.proxy(pf) # for _obtain_enzo
        self.float_type = self.enzo.hierarchy_information["GridLeftEdge"].dtype
        self.directory = os.getcwd()
        AMRHierarchy.__init__(self, pf, data_style)

    def _initialize_data_storage(self):
        pass

    def _count_grids(self):
        self.num_grids = self.enzo.hierarchy_information["GridDimensions"].shape[0]

    def _parse_hierarchy(self):
        self._copy_hierarchy_structure()
        mylog.debug("Copying reverse tree")
        reverse_tree = self.enzo.hierarchy_information["GridParentIDs"].ravel().tolist()
        # Initial setup:
        mylog.debug("Reconstructing parent-child relationships")
        grids = []
        # We enumerate, so it's 0-indexed id and 1-indexed pid
        self.filenames = ["-1"] * self.num_grids
        for id,pid in enumerate(reverse_tree):
            grids.append(self.grid(id+1, self))
            grids[-1].Level = self.grid_levels[id, 0]
            if pid > 0:
                grids[-1]._parent_id = pid
                grids[pid-1]._children_ids.append(grids[-1].id)
        self.max_level = self.grid_levels.max()
        mylog.debug("Preparing grids")
        self.grids = np.empty(len(grids), dtype='object')
        for i, grid in enumerate(grids):
            if (i%1e4) == 0: mylog.debug("Prepared % 7i / % 7i grids", i, self.num_grids)
            grid.filename = None
            grid._prepare_grid()
            grid.proc_num = self.grid_procs[i,0]
            self.grids[i] = grid
        mylog.debug("Prepared")

    def _initialize_grid_arrays(self):
        EnzoHierarchy._initialize_grid_arrays(self)
        self.grid_procs = np.zeros((self.num_grids,1),'int32')

    def _copy_hierarchy_structure(self):
        # Dimensions are important!
        self.grid_dimensions[:] = self.enzo.hierarchy_information["GridEndIndices"][:]
        self.grid_dimensions -= self.enzo.hierarchy_information["GridStartIndices"][:]
        self.grid_dimensions += 1
        self.grid_left_edge[:] = self.enzo.hierarchy_information["GridLeftEdge"][:]
        self.grid_right_edge[:] = self.enzo.hierarchy_information["GridRightEdge"][:]
        self.grid_levels[:] = self.enzo.hierarchy_information["GridLevels"][:]
        self.grid_procs = self.enzo.hierarchy_information["GridProcs"].copy()
        self.grid_particle_count[:] = self.enzo.hierarchy_information["GridNumberOfParticles"][:]

    def save_data(self, *args, **kwargs):
        pass

    _cached_field_list = None
    _cached_derived_field_list = None

    def _detect_fields(self):
        if self.__class__._cached_field_list is None:
            EnzoHierarchy._detect_fields(self)
            self.__class__._cached_field_list = self.field_list
        else:
            self.field_list = self.__class__._cached_field_list

    def _setup_derived_fields(self):
        if self.__class__._cached_derived_field_list is None:
            EnzoHierarchy._setup_derived_fields(self)
            self.__class__._cached_derived_field_list = self.derived_field_list
        else:
            self.derived_field_list = self.__class__._cached_derived_field_list

    def _detect_fields(self):
        self.field_list = []
        # Do this only on the root processor to save disk work.
        mylog.info("Gathering a field list (this may take a moment.)")
        field_list = set()
        random_sample = self._generate_random_grids()
        for grid in random_sample:
            try:
                gf = self.io._read_field_names(grid)
            except self.io._read_exception:
                mylog.debug("Grid %s is a bit funky?", grid.id)
                continue
            mylog.debug("Grid %s has: %s", grid.id, gf)
            field_list = field_list.union(gf)
        field_list = self.comm.par_combine_object(list(field_list),
                        datatype="list", op = "cat")
        self.field_list = list(set(field_list))

    def _generate_random_grids(self):
        my_rank = self.comm.rank
        my_grids = self.grids[self.grid_procs.ravel() == my_rank]
        if len(my_grids) > 40:
            starter = np.random.randint(0, 20)
            random_sample = np.mgrid[starter:len(my_grids)-1:20j].astype("int32")
            mylog.debug("Checking grids: %s", random_sample.tolist())
        else:
            random_sample = np.mgrid[0:max(len(my_grids)-1,1)].astype("int32")
        return my_grids[(random_sample,)]

class EnzoHierarchy1D(EnzoHierarchy):

    def _fill_arrays(self, ei, si, LE, RE, npart):
        self.grid_dimensions[:,:1] = ei
        self.grid_dimensions[:,:1] -= np.array(si, self.float_type)
        self.grid_dimensions += 1
        self.grid_left_edge[:,:1] = LE
        self.grid_right_edge[:,:1] = RE
        self.grid_particle_count.flat[:] = npart
        self.grid_left_edge[:,1:] = 0.0
        self.grid_right_edge[:,1:] = 1.0
        self.grid_dimensions[:,1:] = 1

class EnzoHierarchy2D(EnzoHierarchy):

    def _fill_arrays(self, ei, si, LE, RE, npart):
        self.grid_dimensions[:,:2] = ei
        self.grid_dimensions[:,:2] -= np.array(si, self.float_type)
        self.grid_dimensions += 1
        self.grid_left_edge[:,:2] = LE
        self.grid_right_edge[:,:2] = RE
        self.grid_particle_count.flat[:] = npart
        self.grid_left_edge[:,2] = 0.0
        self.grid_right_edge[:,2] = 1.0
        self.grid_dimensions[:,2] = 1

class EnzoStaticOutput(StaticOutput):
    """
    Enzo-specific output, set at a fixed time.
    """
    _hierarchy_class = EnzoHierarchy
    _fieldinfo_fallback = EnzoFieldInfo
    _fieldinfo_known = KnownEnzoFields
    def __init__(self, filename, data_style=None,
                 file_style = None,
                 parameter_override = None,
                 conversion_override = None,
                 storage_filename = None):
        """
        This class is a stripped down class that simply reads and parses
        *filename* without looking at the hierarchy.  *data_style* gets passed
        to the hierarchy to pre-determine the style of data-output.  However,
        it is not strictly necessary.  Optionally you may specify a
        *parameter_override* dictionary that will override anything in the
        paarmeter file and a *conversion_override* dictionary that consists
        of {fieldname : conversion_to_cgs} that will override the #DataCGS.
        """
        if filename.endswith(".hierarchy"): filename = filename[:-10]
        if parameter_override is None: parameter_override = {}
        self._parameter_override = parameter_override
        if conversion_override is None: conversion_override = {}
        self._conversion_override = conversion_override
        self.storage_filename = storage_filename

        StaticOutput.__init__(self, filename, data_style, file_style=file_style)
        if "InitialTime" not in self.parameters:
            self.current_time = 0.0

    def _setup_1d(self):
        self._hierarchy_class = EnzoHierarchy1D
        self._fieldinfo_fallback = Enzo1DFieldInfo
        self.domain_left_edge = \
            np.concatenate([[self.domain_left_edge], [0.0, 0.0]])
        self.domain_right_edge = \
            np.concatenate([[self.domain_right_edge], [1.0, 1.0]])

    def _setup_2d(self):
        self._hierarchy_class = EnzoHierarchy2D
        self._fieldinfo_fallback = Enzo2DFieldInfo
        self.domain_left_edge = \
            np.concatenate([self.domain_left_edge, [0.0]])
        self.domain_right_edge = \
            np.concatenate([self.domain_right_edge, [1.0]])

    def get_parameter(self,parameter,type=None):
        """
        Gets a parameter not in the parameterDict.
        """
        if self.parameters.has_key(parameter):
            return self.parameters[parameter]

        # Let's read the file
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])
        lines = open(self.parameter_filename).readlines()
        for lineI, line in enumerate(lines):
            if line.find("#") >= 1: # Keep the commented lines
                line=line[:line.find("#")]
            line=line.strip().rstrip()
            if len(line) < 2:
                continue
            try:
                param, vals = map(string.strip,map(string.rstrip,
                                                   line.split("=")))
            except ValueError:
                mylog.error("ValueError: '%s'", line)
            if parameter == param:
                if type is None:
                    t = vals.split()
                else:
                    t = map(type, vals.split())
                if len(t) == 1:
                    self.parameters[param] = t[0]
                else:
                    self.parameters[param] = t
                if param.endswith("Units") and not param.startswith("Temperature"):
                    dataType = param[:-5]
                    self.conversion_factors[dataType] = self.parameters[param]
                return self.parameters[parameter]

        return ""

    def _parse_parameter_file(self):
        """
        Parses the parameter file and establishes the various
        dictionaries.
        """
        # Let's read the file
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])
        lines = open(self.parameter_filename).readlines()
        data_labels = {}
        data_label_factors = {}
        for line in (l.strip() for l in lines):
            if len(line) < 2: continue
            param, vals = (i.strip() for i in line.split("=",1))
            # First we try to decipher what type of value it is.
            vals = vals.split()
            # Special case approaching.
            if "(do" in vals: vals = vals[:1]
            if len(vals) == 0:
                pcast = str # Assume NULL output
            else:
                v = vals[0]
                # Figure out if it's castable to floating point:
                try:
                    float(v)
                except ValueError:
                    pcast = str
                else:
                    if any("." in v or "e+" in v or "e-" in v for v in vals):
                        pcast = float
                    elif v == "inf":
                        pcast = str
                    else:
                        pcast = int
            # Now we figure out what to do with it.
            if param.endswith("Units") and not param.startswith("Temperature"):
                dataType = param[:-5]
                # This one better be a float.
                self.conversion_factors[dataType] = float(vals[0])
            if param.startswith("#DataCGS") or \
                 param.startswith("#CGSConversionFactor"):
                # Assume of the form: #DataCGSConversionFactor[7] = 2.38599e-26 g/cm^3
                # Which one does it belong to?
                data_id = param[param.find("[")+1:param.find("]")]
                data_label_factors[data_id] = float(vals[0])
            if param.startswith("DataLabel"):
                data_id = param[param.find("[")+1:param.find("]")]
                data_labels[data_id] = vals[0]
            if len(vals) == 0:
                vals = ""
            elif len(vals) == 1:
                vals = pcast(vals[0])
            else:
                vals = np.array([pcast(i) for i in vals if i != "-99999"])
            self.parameters[param] = vals
        for p, v in self._parameter_override.items():
            self.parameters[p] = v
        for p, v in self._conversion_override.items():
            self.conversion_factors[p] = v
        for k, v in data_label_factors.items():
            self.conversion_factors[data_labels[k]] = v
        self.refine_by = self.parameters["RefineBy"]
        self.periodicity = ensure_tuple(self.parameters["LeftFaceBoundaryCondition"] == 3)
        self.dimensionality = self.parameters["TopGridRank"]
        if self.dimensionality > 1:
            self.domain_dimensions = self.parameters["TopGridDimensions"]
            if len(self.domain_dimensions) < 3:
                tmp = self.domain_dimensions.tolist()
                tmp.append(1)
                self.domain_dimensions = np.array(tmp)
                self.periodicity += (False,)
            self.domain_left_edge = np.array(self.parameters["DomainLeftEdge"],
                                             "float64").copy()
            self.domain_right_edge = np.array(self.parameters["DomainRightEdge"],
                                             "float64").copy()
        else:
            self.domain_left_edge = np.array(self.parameters["DomainLeftEdge"],
                                             "float64")
            self.domain_right_edge = np.array(self.parameters["DomainRightEdge"],
                                             "float64")
            self.domain_dimensions = np.array([self.parameters["TopGridDimensions"],1,1])
            self.periodicity += (False, False)

        self.current_time = self.parameters["InitialTime"]
        # To be enabled when we can break old pickles:
        #if "MetaDataSimulationUUID" in self.parameters:
        #    self.unique_identifier = self.parameters["MetaDataSimulationUUID"]
        if "CurrentTimeIdentifier" in self.parameters:
            self.unique_identifier = self.parameters["CurrentTimeIdentifier"]
        if self.parameters["ComovingCoordinates"]:
            self.cosmological_simulation = 1
            self.current_redshift = self.parameters["CosmologyCurrentRedshift"]
            self.omega_lambda = self.parameters["CosmologyOmegaLambdaNow"]
            self.omega_matter = self.parameters["CosmologyOmegaMatterNow"]
            self.hubble_constant = self.parameters["CosmologyHubbleConstantNow"]
        else:
            self.current_redshift = self.omega_lambda = self.omega_matter = \
                self.hubble_constant = self.cosmological_simulation = 0.0

        if self.dimensionality == 1:
            self._setup_1d()
        elif self.dimensionality == 2:
            self._setup_2d()

    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        if "EOSType" not in self.parameters: self.parameters["EOSType"] = -1
        if self["ComovingCoordinates"]:
            self._setup_comoving_units()
        elif self.has_key("LengthUnit"):
            # 'Why share when we can reinvent incompatibly?'
            self.parameters["LengthUnits"] = self["LengthUnit"]
            self._setup_getunits_units()
        elif self.has_key("LengthUnits"):
            self._setup_getunits_units()
        else:
            self._setup_nounits_units()
        self.time_units['1'] = 1
        self.units['1'] = 1
        self.units['unitary'] = 1.0 / (self.domain_right_edge - self.domain_left_edge).max()
        for unit in sec_conversion.keys():
            self.time_units[unit] = self["Time"] / sec_conversion[unit]

    def _setup_comoving_units(self):
        z = self["CosmologyCurrentRedshift"]
        h = self["CosmologyHubbleConstantNow"]
        boxcm_cal = self["CosmologyComovingBoxSize"]
        boxcm_uncal = boxcm_cal / h
        box_proper = boxcm_uncal/(1+z)
        self.units['aye']  = (1.0 + self["CosmologyInitialRedshift"])/(z + 1.0)
        if not self.has_key("Time"):
            cu = self.cosmology_get_units()
            self.conversion_factors["Time"] = cu['utim']
        for unit in mpc_conversion:
            self.units[unit] = mpc_conversion[unit] * box_proper
            self.units[unit+'h'] = mpc_conversion[unit] * box_proper * h
            self.units[unit+'cm'] = mpc_conversion[unit] * boxcm_uncal
            self.units[unit+'hcm'] = mpc_conversion[unit] * boxcm_cal

    def _setup_getunits_units(self):
        # We are given LengthUnits, which is number of cm per box length
        # So we convert that to box-size in Mpc
        box_proper = 3.24077e-25 * self["LengthUnits"]
        self.units['aye']  = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] * box_proper
        if not self.has_key("TimeUnits"):
            self.conversion_factors["Time"] = self["LengthUnits"] / self["x-velocity"]

    def _setup_nounits_units(self):
        z = 0
        mylog.warning("Setting 1.0 in code units to be 1.0 cm")
        if not self.has_key("TimeUnits"):
            mylog.warning("No time units.  Setting 1.0 = 1 second.")
            self.conversion_factors["Time"] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]

    def cosmology_get_units(self):
        """
        Return an Enzo-fortran style dictionary of units to feed into custom
        routines.  This is typically only necessary if you are interacting
        with fortran code.
        """
        k = {}
        k["utim"] = 2.52e17/np.sqrt(self.omega_matter)\
                       / self.hubble_constant \
                       / (1+self.parameters["CosmologyInitialRedshift"])**1.5
        k["urho"] = 1.88e-29 * self.omega_matter \
                        * self.hubble_constant**2 \
                        * (1.0 + self.current_redshift)**3
        k["uxyz"] = 3.086e24 * \
               self.parameters["CosmologyComovingBoxSize"] / \
               self.hubble_constant / \
               (1.0 + self.current_redshift)
        k["uaye"] = 1.0/(1.0 + self.parameters["CosmologyInitialRedshift"])
        k["uvel"] = 1.225e7*self.parameters["CosmologyComovingBoxSize"] \
                      *np.sqrt(self.omega_matter) \
                      *np.sqrt(1+ self.parameters["CosmologyInitialRedshift"])
        k["utem"] = 1.88e6 * (self.parameters["CosmologyComovingBoxSize"]**2) \
                      * self.omega_matter \
                      * (1.0 + self.parameters["CosmologyInitialRedshift"])
        k["aye"]  = (1.0 + self.parameters["CosmologyInitialRedshift"]) / \
               (1.0 + self.current_redshift)
        return k

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        if ("%s" % (args[0])).endswith(".hierarchy"):
            return True
        return os.path.exists("%s.hierarchy" % args[0])

class EnzoStaticOutputInMemory(EnzoStaticOutput):
    _hierarchy_class = EnzoHierarchyInMemory
    _data_style = 'enzo_inline'

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj

    def __init__(self, parameter_override=None, conversion_override=None):
        if parameter_override is None: parameter_override = {}
        self._parameter_override = parameter_override
        if conversion_override is None: conversion_override = {}
        self._conversion_override = conversion_override

        StaticOutput.__init__(self, "InMemoryParameterFile", self._data_style)

    def _parse_parameter_file(self):
        enzo = self._obtain_enzo()
        self.basename = "cycle%08i" % (
            enzo.yt_parameter_file["NumberOfPythonCalls"])
        self.parameters['CurrentTimeIdentifier'] = time.time()
        self.parameters.update(enzo.yt_parameter_file)
        self.conversion_factors.update(enzo.conversion_factors)
        for i in self.parameters:
            if isinstance(self.parameters[i], types.TupleType):
                self.parameters[i] = np.array(self.parameters[i])
            if i.endswith("Units") and not i.startswith("Temperature"):
                dataType = i[:-5]
                self.conversion_factors[dataType] = self.parameters[i]
        self.domain_left_edge = self.parameters["DomainLeftEdge"].copy()
        self.domain_right_edge = self.parameters["DomainRightEdge"].copy()
        for i in self.conversion_factors:
            if isinstance(self.conversion_factors[i], types.TupleType):
                self.conversion_factors[i] = np.array(self.conversion_factors[i])
        for p, v in self._parameter_override.items():
            self.parameters[p] = v
        for p, v in self._conversion_override.items():
            self.conversion_factors[p] = v
        self.refine_by = self.parameters["RefineBy"]
        self.periodicity = ensure_tuple(self.parameters["LeftFaceBoundaryCondition"] == 3)
        self.dimensionality = self.parameters["TopGridRank"]
        self.domain_dimensions = self.parameters["TopGridDimensions"]
        self.current_time = self.parameters["InitialTime"]
        if "CurrentTimeIdentifier" in self.parameters:
            self.unique_identifier = self.parameters["CurrentTimeIdentifier"]
        if self.parameters["ComovingCoordinates"]:
            self.cosmological_simulation = 1
            self.current_redshift = self.parameters["CosmologyCurrentRedshift"]
            self.omega_lambda = self.parameters["CosmologyOmegaLambdaNow"]
            self.omega_matter = self.parameters["CosmologyOmegaMatterNow"]
            self.hubble_constant = self.parameters["CosmologyHubbleConstantNow"]
        else:
            self.current_redshift = self.omega_lambda = self.omega_matter = \
                self.hubble_constant = self.cosmological_simulation = 0.0

    def _obtain_enzo(self):
        import enzo; return enzo

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        return False

# These next two functions are taken from
# http://www.reddit.com/r/Python/comments/6hj75/reverse_file_iterator/c03vms4
# Credit goes to "Brian" on Reddit

def rblocks(f, blocksize=4096):
    """Read file as series of blocks from end of file to start.

    The data itself is in normal order, only the order of the blocks is reversed.
    ie. "hello world" -> ["ld","wor", "lo ", "hel"]
    Note that the file must be opened in binary mode.
    """
    if 'b' not in f.mode.lower():
        raise Exception("File must be opened using binary mode.")
    size = os.stat(f.name).st_size
    fullblocks, lastblock = divmod(size, blocksize)

    # The first(end of file) block will be short, since this leaves 
    # the rest aligned on a blocksize boundary.  This may be more 
    # efficient than having the last (first in file) block be short
    f.seek(-lastblock,2)
    yield f.read(lastblock)

    for i in range(fullblocks-1,-1, -1):
        f.seek(i * blocksize)
        yield f.read(blocksize)

def rlines(f, keepends=False):
    """Iterate through the lines of a file in reverse order.

    If keepends is true, line endings are kept as part of the line.
    """
    buf = ''
    for block in rblocks(f):
        buf = block + buf
        lines = buf.splitlines(keepends)
        # Return all lines except the first (since may be partial)
        if lines:
            lines.reverse()
            buf = lines.pop() # Last line becomes end of new first line.
            for line in lines:
                yield line
    yield buf  # First line.




"""
A two-pass contour finding algorithm

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free

cdef extern from "math.h":
    double fabs(double x)

cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1):
    if i0 > i1: return i0
    return i1

cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1):
    if i0 < i1: return i0
    return i1

cdef extern from "union_find.h":
    ctypedef struct forest_node:
        void *value
        forest_node *parent
        int rank

    forest_node* MakeSet(void* value)
    void Union(forest_node* node1, forest_node* node2)
    forest_node* Find(forest_node* node)

ctypedef struct CellIdentifier:
    np.int64_t hindex
    int level

cdef class GridContourContainer:
    cdef np.int64_t dims[3]
    cdef np.int64_t start_indices[3]
    cdef forest_node **join_tree
    cdef np.int64_t ncells

    def __init__(self, dimensions, indices):
        cdef int i
        self.ncells = 1
        for i in range(3):
            self.ncells *= dimensions[i]
            self.dims[i] = dimensions[i]
            self.start_indices[i] = indices[i]
        self.join_tree = <forest_node **> malloc(sizeof(forest_node) 
                                                 * self.ncells)
        for i in range(self.ncells): self.join_tree[i] = NULL

    def __dealloc__(self):
        cdef int i
        for i in range(self.ncells):
            if self.join_tree[i] != NULL: free(self.join_tree[i])
        free(self.join_tree)

    #def construct_join_tree(self,
    #            np.ndarray[np.float64_t, ndim=3] field,
    #            np.ndarray[np.bool_t, ndim=3] mask):
    #    # This only looks at the components of the grid that are actually
    #    # inside this grid -- boundary conditions are handled later.
    #    pass

#@cython.boundscheck(False)
#@cython.wraparound(False)
def construct_boundary_relationships(
        np.ndarray[dtype=np.int64_t, ndim=3] contour_ids):
    # We only look at the boundary and one cell in
    cdef int i, j, nx, ny, nz, offset_i, offset_j, oi, oj
    cdef np.int64_t c1, c2
    nx = contour_ids.shape[0]
    ny = contour_ids.shape[1]
    nz = contour_ids.shape[2]
    # We allocate an array of fixed (maximum) size
    cdef int s = (ny*nx + nx*nz + ny*nz - 2) * 18
    cdef np.ndarray[np.int64_t, ndim=2] tree = np.zeros((s, 2), dtype="int64")
    cdef int ti = 0
    # First x-pass
    for i in range(ny):
        for j in range(nz):
            for offset_i in range(3):
                oi = offset_i - 1
                if i == 0 and oi == -1: continue
                if i == ny - 1 and oi == 1: continue
                for offset_j in range(3):
                    oj = offset_j - 1
                    if j == 0 and oj == -1: continue
                    if j == nz - 1 and oj == 1: continue
                    c1 = contour_ids[0, i, j]
                    c2 = contour_ids[1, i + oi, j + oj]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
                    c1 = contour_ids[nx-1, i, j]
                    c2 = contour_ids[nx-2, i + oi, j + oj]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
    # Now y-pass
    for i in range(nx):
        for j in range(nz):
            for offset_i in range(3):
                oi = offset_i - 1
                if i == 0 and oi == -1: continue
                if i == nx - 1 and oi == 1: continue
                for offset_j in range(3):
                    oj = offset_j - 1
                    if j == 0 and oj == -1: continue
                    if j == nz - 1 and oj == 1: continue
                    c1 = contour_ids[i, 0, j]
                    c2 = contour_ids[i + oi, 1, j + oj]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
                    c1 = contour_ids[i, ny-1, j]
                    c2 = contour_ids[i + oi, ny-2, j + oj]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
    for i in range(nx):
        for j in range(ny):
            for offset_i in range(3):
                oi = offset_i - 1
                if i == 0 and oi == -1: continue
                if i == nx - 1 and oi == 1: continue
                for offset_j in range(3):
                    oj = offset_j - 1
                    if j == 0 and oj == -1: continue
                    if j == ny - 1 and oj == 1: continue
                    c1 = contour_ids[i, j, 0]
                    c2 = contour_ids[i + oi, j + oj, 1]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
                    c1 = contour_ids[i, j, nz-1]
                    c2 = contour_ids[i + oi, j + oj, nz-2]
                    if c1 > -1 and c2 > -1:
                        tree[ti,0] = i64max(c1,c2)
                        tree[ti,1] = i64min(c1,c2)
                        ti += 1
    return tree[:ti,:]

cdef inline int are_neighbors(
            np.float64_t x1, np.float64_t y1, np.float64_t z1,
            np.float64_t dx1, np.float64_t dy1, np.float64_t dz1,
            np.float64_t x2, np.float64_t y2, np.float64_t z2,
            np.float64_t dx2, np.float64_t dy2, np.float64_t dz2,
        ):
    # We assume an epsilon of 1e-15
    if fabs(x1-x2) > 0.5*(dx1+dx2): return 0
    if fabs(y1-y2) > 0.5*(dy1+dy2): return 0
    if fabs(z1-z2) > 0.5*(dz1+dz2): return 0
    return 1

@cython.boundscheck(False)
@cython.wraparound(False)
def identify_field_neighbors(
            np.ndarray[dtype=np.float64_t, ndim=1] field,
            np.ndarray[dtype=np.float64_t, ndim=1] x,
            np.ndarray[dtype=np.float64_t, ndim=1] y,
            np.ndarray[dtype=np.float64_t, ndim=1] z,
            np.ndarray[dtype=np.float64_t, ndim=1] dx,
            np.ndarray[dtype=np.float64_t, ndim=1] dy,
            np.ndarray[dtype=np.float64_t, ndim=1] dz,
        ):
    # We assume this field is pre-jittered; it has no identical values.
    cdef int outer, inner, N, added
    cdef np.float64_t x1, y1, z1, dx1, dy1, dz1
    N = field.shape[0]
    #cdef np.ndarray[dtype=np.object_t] joins
    joins = [[] for outer in range(N)]
    #joins = np.empty(N, dtype='object')
    for outer in range(N):
        if (outer % 10000) == 0: print outer, N
        x1 = x[outer]
        y1 = y[outer]
        z1 = z[outer]
        dx1 = dx[outer]
        dy1 = dy[outer]
        dz1 = dz[outer]
        this_joins = joins[outer]
        added = 0
        # Go in reverse order
        for inner in range(outer, 0, -1):
            if not are_neighbors(x1, y1, z1, dx1, dy1, dz1,
                                 x[inner], y[inner], z[inner],
                                 dx[inner], dy[inner], dz[inner]):
                continue
            # Hot dog, we have a weiner!
            this_joins.append(inner)
            added += 1
            if added == 26: break
    return joins

@cython.boundscheck(False)
@cython.wraparound(False)
def extract_identified_contours(int max_ind, joins):
    cdef int i
    contours = []
    for i in range(max_ind + 1): # +1 to get to the max_ind itself
        contours.append(set([i]))
        if len(joins[i]) == 0:
            continue
        proto_contour = [i]
        for j in joins[i]:
            proto_contour += contours[j]
        proto_contour = set(proto_contour)
        for j in proto_contour:
            contours[j] = proto_contour
    return contours

@cython.boundscheck(False)
@cython.wraparound(False)
def update_joins(joins, np.ndarray[np.int64_t, ndim=1] contour_ids):
    cdef np.int64_t new, old, i, oi
    cdef int n, on
    cdef np.ndarray[np.int64_t, ndim=1] old_set
    #print contour_ids.shape[0]
    n = contour_ids.shape[0]
    for new, old_set in joins:
        #print new
        on = old_set.shape[0]
        for i in range(n):
            for oi in range(on):
                old = old_set[oi]
                if contour_ids[i] == old: contour_ids[i] = new



"""
Integrator classes to deal with interpolation and integration of input spectral
bins.  Currently only supports Cloudy-style data.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: Michigan State University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2012 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from exceptions import IOError
import h5py
import numpy as np
import os

from yt.funcs import *

from yt.data_objects.field_info_container import add_field
from yt.utilities.exceptions import YTException
from yt.utilities.linear_interpolators import \
    BilinearFieldInterpolator

xray_data_version = 1

class EnergyBoundsException(YTException):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper

    def __str__(self):
        return "Energy bounds are %e to %e keV." % \
          (self.lower, self.upper)

class ObsoleteDataException(YTException):
    def __str__(self):
        return "X-ray emissivity data is out of data.\nDownload the latest data from http://yt-project.org/data/xray_emissivity.h5 and move it to %s." % \
          os.path.join(os.environ["YT_DEST"], "data", "xray_emissivity.h5")
          
class EmissivityIntegrator(object):
    r"""Class for making X-ray emissivity fields with hdf5 data tables 
    from Cloudy.
    
    Initialize an EmissivityIntegrator object.

    Parameters
    ----------
    filename: string, default None
        Path to data file containing emissivity values.  If None,
        a file called xray_emissivity.h5 is used.  This file contains 
        emissivity tables for primordial elements and for metals at 
        solar metallicity for the energy range 0.1 to 100 keV.
        Default: None.
        
    """
    def __init__(self, filename=None):

        default_filename = False
        if filename is None:
            filename = os.path.join(os.environ["YT_DEST"], 
                                    "data", "xray_emissivity.h5")
            default_filename = True

        if not os.path.exists(filename):
            raise IOError("File does not exist: %s." % filename)
        only_on_root(mylog.info, "Loading emissivity data from %s." % filename)
        in_file = h5py.File(filename, "r")
        if "info" in in_file.attrs:
            only_on_root(mylog.info, in_file.attrs["info"])
        if default_filename and \
          in_file.attrs["version"] < xray_data_version:
            raise ObsoleteDataException()
        else:
            only_on_root(mylog.info, "X-ray emissivity data version: %s." % \
                         in_file.attrs["version"])

        for field in ["emissivity_primordial", "emissivity_metals",
                      "log_nH", "log_T", "log_E"]:
            setattr(self, field, in_file[field][:])
        in_file.close()

        E_diff = np.diff(self.log_E)
        self.E_bins = \
                  np.power(10, np.concatenate([self.log_E[:-1] - 0.5 * E_diff,
                                               [self.log_E[-1] - 0.5 * E_diff[-1],
                                                self.log_E[-1] + 0.5 * E_diff[-1]]]))
        self.dnu = 2.41799e17 * np.diff(self.E_bins)

    def _get_interpolator(self, data, e_min, e_max):
        r"""Create an interpolator for total emissivity in a 
        given energy range.

        Parameters
        ----------
        e_min: float
            the minimum energy in keV for the energy band.
        e_min: float
            the maximum energy in keV for the energy band.

        """
        if (e_min - self.E_bins[0]) / e_min < -1e-3 or \
          (e_max - self.E_bins[-1]) / e_max > 1e-3:
            raise EnergyBoundsException(np.power(10, self.E_bins[0]),
                                        np.power(10, self.E_bins[-1]))
        e_is, e_ie = np.digitize([e_min, e_max], self.E_bins)
        e_is = np.clip(e_is - 1, 0, self.E_bins.size - 1)
        e_ie = np.clip(e_ie, 0, self.E_bins.size - 1)

        my_dnu = np.copy(self.dnu[e_is: e_ie])
        # clip edge bins if the requested range is smaller
        my_dnu[0] -= e_min - self.E_bins[e_is]
        my_dnu[-1] -= self.E_bins[e_ie] - e_max

        interp_data = (data[..., e_is:e_ie] * my_dnu).sum(axis=-1)
        return BilinearFieldInterpolator(np.log10(interp_data),
                                         [self.log_nH[0], self.log_nH[-1],
                                          self.log_T[0],  self.log_T[-1]],
                                         ["log_nH", "log_T"], truncate=True)

def add_xray_emissivity_field(e_min, e_max, filename=None,
                              with_metals=True,
                              constant_metallicity=None):
    r"""Create an X-ray emissivity field for a given energy range.

    Parameters
    ----------
    e_min: float
        the minimum energy in keV for the energy band.
    e_min: float
        the maximum energy in keV for the energy band.

    Other Parameters
    ----------------
    filename: string
        Path to data file containing emissivity values.  If None,
        a file called xray_emissivity.h5 is used.  This file contains 
        emissivity tables for primordial elements and for metals at 
        solar metallicity for the energy range 0.1 to 100 keV.
        Default: None.
    with_metals: bool
        If True, use the metallicity field to add the contribution from 
        metals.  If False, only the emission from H/He is considered.
        Default: True.
    constant_metallicity: float
        If specified, assume a constant metallicity for the emission 
        from metals.  The *with_metals* keyword must be set to False 
        to use this.
        Default: None.

    This will create a field named "Xray_Emissivity_{e_min}_{e_max}keV".
    The units of the field are erg s^-1 cm^-3.

    Examples
    --------

    >>> from yt.mods import *
    >>> from yt.analysis_modules.spectral_integrator.api import *
    >>> add_xray_emissivity_field(0.5, 2)
    >>> pf = load(dataset)
    >>> p = ProjectionPlot(pf, 'x', "Xray_Emissivity_0.5_2keV")
    >>> p.save()

    """

    my_si = EmissivityIntegrator(filename=filename)

    em_0 = my_si._get_interpolator(my_si.emissivity_primordial, e_min, e_max)
    em_Z = None
    if with_metals or constant_metallicity is not None:
        em_Z = my_si._get_interpolator(my_si.emissivity_metals, e_min, e_max)

    def _emissivity_field(field, data):
        dd = {"log_nH" : np.log10(data["H_NumberDensity"]),
              "log_T"   : np.log10(data["Temperature"])}

        my_emissivity = np.power(10, em_0(dd))
        if em_Z is not None:
            if with_metals:
                my_Z = data["Metallicity"]
            elif constant_metallicity is not None:
                my_Z = constant_metallicity
            my_emissivity += my_Z * np.power(10, em_Z(dd))

        return data["H_NumberDensity"]**2 * my_emissivity

    field_name = "Xray_Emissivity_%s_%skeV" % (e_min, e_max)
    add_field(field_name, function=_emissivity_field,
              projection_conversion="cm",
              display_name=r"\epsilon_{X}\/(%s-%s\/keV)" % (e_min, e_max),
              units=r"\rm{erg}\/\rm{cm}^{-3}\/\rm{s}^{-1}",
              projected_units=r"\rm{erg}\/\rm{cm}^{-2}\/\rm{s}^{-1}")
    return field_name

def add_xray_luminosity_field(e_min, e_max, filename=None,
                              with_metals=True,
                              constant_metallicity=None):
    r"""Create an X-ray luminosity field for a given energy range.

    Parameters
    ----------
    e_min: float
        the minimum energy in keV for the energy band.
    e_min: float
        the maximum energy in keV for the energy band.

    Other Parameters
    ----------------
    filename: string
        Path to data file containing emissivity values.  If None,
        a file called xray_emissivity.h5 is used.  This file contains 
        emissivity tables for primordial elements and for metals at 
        solar metallicity for the energy range 0.1 to 100 keV.
        Default: None.
    with_metals: bool
        If True, use the metallicity field to add the contribution from 
        metals.  If False, only the emission from H/He is considered.
        Default: True.
    constant_metallicity: float
        If specified, assume a constant metallicity for the emission 
        from metals.  The *with_metals* keyword must be set to False 
        to use this.
        Default: None.

    This will create a field named "Xray_Luminosity_{e_min}_{e_max}keV".
    The units of the field are erg s^-1.

    Examples
    --------

    >>> from yt.mods import *
    >>> from yt.analysis_modules.spectral_integrator.api import *
    >>> add_xray_luminosity_field(0.5, 2)
    >>> pf = load(dataset)
    >>> sp = pf.h.sphere('max', (2., 'mpc'))
    >>> print sp.quantities['TotalQuantity']('Xray_Luminosity_0.5_2keV')
    
    """

    em_field = add_xray_emissivity_field(e_min, e_max, filename=filename,
                                         with_metals=with_metals,
                                         constant_metallicity=constant_metallicity)

    def _luminosity_field(field, data):
        return data[em_field] * data["CellVolume"]
    field_name = "Xray_Luminosity_%s_%skeV" % (e_min, e_max)
    add_field(field_name, function=_luminosity_field,
              display_name=r"\rm{L}_{X}\/(%s-%s\/keV)" % (e_min, e_max),
              units=r"\rm{erg}\/\rm{s}^{-1}")
    return field_name

def add_xray_photon_emissivity_field(e_min, e_max, filename=None,
                                     with_metals=True,
                                     constant_metallicity=None):
    r"""Create an X-ray photon emissivity field for a given energy range.

    Parameters
    ----------
    e_min: float
        the minimum energy in keV for the energy band.
    e_min: float
        the maximum energy in keV for the energy band.

    Other Parameters
    ----------------
    filename: string
        Path to data file containing emissivity values.  If None,
        a file called xray_emissivity.h5 is used.  This file contains 
        emissivity tables for primordial elements and for metals at 
        solar metallicity for the energy range 0.1 to 100 keV.
        Default: None.
    with_metals: bool
        If True, use the metallicity field to add the contribution from 
        metals.  If False, only the emission from H/He is considered.
        Default: True.
    constant_metallicity: float
        If specified, assume a constant metallicity for the emission 
        from metals.  The *with_metals* keyword must be set to False 
        to use this.
        Default: None.

    This will create a field named "Xray_Photon_Emissivity_{e_min}_{e_max}keV".
    The units of the field are photons s^-1 cm^-3.

    Examples
    --------

    >>> from yt.mods import *
    >>> from yt.analysis_modules.spectral_integrator.api import *
    >>> add_xray_emissivity_field(0.5, 2)
    >>> pf = load(dataset)
    >>> p = ProjectionPlot(pf, 'x', "Xray_Emissivity_0.5_2keV")
    >>> p.save()

    """

    my_si = EmissivityIntegrator(filename=filename)
    energy_erg = np.power(10, my_si.log_E) * 1.60217646e-9

    em_0 = my_si._get_interpolator((my_si.emissivity_primordial[..., :] / energy_erg),
                                   e_min, e_max)
    em_Z = None
    if with_metals or constant_metallicity is not None:
        em_Z = my_si._get_interpolator((my_si.emissivity_metals[..., :] / energy_erg),
                                       e_min, e_max)

    def _emissivity_field(field, data):
        dd = {"log_nH" : np.log10(data["H_NumberDensity"]),
              "log_T"   : np.log10(data["Temperature"])}

        my_emissivity = np.power(10, em_0(dd))
        if em_Z is not None:
            if with_metals:
                my_Z = data["Metallicity"]
            elif constant_metallicity is not None:
                my_Z = constant_metallicity
            my_emissivity += my_Z * np.power(10, em_Z(dd))

        return data["H_NumberDensity"]**2 * my_emissivity

    field_name = "Xray_Photon_Emissivity_%s_%skeV" % (e_min, e_max)
    add_field(field_name, function=_emissivity_field,
              projection_conversion="cm",
              display_name=r"\epsilon_{X}\/(%s-%s\/keV)" % (e_min, e_max),
              units=r"\rm{photons}\/\rm{cm}^{-3}\/\rm{s}^{-1}",
              projected_units=r"\rm{photons}\/\rm{cm}^{-2}\/\rm{s}^{-1}")
    return field_name



"""
Writing yt data to a GDF file.

Authors: Casey W. Stark <caseywstark@gmail.com>
Affiliation: UC Berkeley

Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Casey W. Stark.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
import sys
import h5py
import numpy as np

from yt import __version__ as yt_version


def write_to_gdf(pf, gdf_path, data_author=None, data_comment=None,
                 particle_type_name="dark_matter"):
    """
    Write a parameter file to the given path in the Grid Data Format.

    Parameters
    ----------
    pf : StaticOutput object
        The yt data to write out.
    gdf_path : string
        The path of the file to output.

    """

    f = _create_new_gdf(pf, gdf_path, data_author, data_comment,
                       particle_type_name)

    # now add the fields one-by-one
    for field_name in pf.h.field_list:
        _write_field_to_gdf(pf, f, field_name, particle_type_name)

    # don't forget to close the file.
    f.close()

def save_field(pf, field_name, field_parameters=None):
    """
    Write a single field associated with the parameter file pf to the
    backup file.

    Parameters
    ----------
    pf : StaticOutput object
        The yt parameter file that the field is associated with.
    field_name : string
        The name of the field to save.
    field_parameters : dictionary
        A dictionary of field parameters to set.
    """

    field_obj = pf.field_info[field_name]
    if field_obj.particle_type:
        print( "Saving particle fields currently not supported." )
        return
    
    backup_filename = pf.backup_filename
    if os.path.exists(backup_filename):
        # backup file already exists, open it
        f = h5py.File(backup_filename, "r+")
    else:
        # backup file does not exist, create it
        f = _create_new_gdf(pf, backup_filename, data_author=None, data_comment=None,
                       particle_type_name="dark_matter")

    # now save the field
    _write_field_to_gdf(pf, f, field_name, particle_type_name="dark_matter", field_parameters=field_parameters)

    # don't forget to close the file.
    f.close()
        
def _write_field_to_gdf(pf, fhandle, field_name, particle_type_name, field_parameters=None):

    # add field info to field_types group
    g = fhandle["field_types"]
    # create the subgroup with the field's name
    try:
        sg = g.create_group(field_name)
    except ValueError:
        print "Error - File already contains field called " + field_name
        sys.exit(1)
        
    # grab the display name and units from the field info container.
    display_name = pf.field_info[field_name].display_name
    units = pf.field_info[field_name].get_units()

    # check that they actually contain something...
    if display_name:
        sg.attrs["field_name"] = display_name
    else:
        sg.attrs["field_name"] = field_name
    if units:
        sg.attrs["field_units"] = units
    else:
        sg.attrs["field_units"] = "None"
    # @todo: the values must be in CGS already right?
    sg.attrs["field_to_cgs"] = 1.0
    # @todo: is this always true?
    sg.attrs["staggering"] = 0

    # now add actual data, grid by grid
    g = fhandle["data"]     
    for grid in pf.h.grids:

        # set field parameters, if specified
        if field_parameters is not None:
            for k,v in field_parameters.iteritems():
                grid.set_field_parameter(k,v)

        grid_group = g["grid_%010i" % (grid.id - grid._id_offset)]
        particles_group = grid_group["particles"]
        pt_group = particles_group[particle_type_name]
        # add the field data to the grid group
        # Check if this is a real field or particle data.
        field_obj = pf.field_info[field_name]
        if field_obj.particle_type:  # particle data
            pt_group[field_name] = grid.get_data(field_name)
        else:  # a field
            grid_group[field_name] = grid.get_data(field_name)

def _create_new_gdf(pf, gdf_path, data_author=None, data_comment=None,
                   particle_type_name="dark_matter"):
    # Make sure we have the absolute path to the file first
    gdf_path = os.path.abspath(gdf_path)

    # Stupid check -- is the file already there?
    # @todo: make this a specific exception/error.
    if os.path.exists(gdf_path):
        raise IOError("A file already exists in the location: %s. Please provide a new one or remove that file." % gdf_path)

    ###
    # Create and open the file with h5py
    ###
    f = h5py.File(gdf_path, "w")

    ###
    # "gridded_data_format" group
    ###
    g = f.create_group("gridded_data_format")
    g.attrs["data_software"] = "yt"
    g.attrs["data_software_version"] = yt_version
    if data_author is not None:
        g.attrs["data_author"] = data_author
    if data_comment is not None:
        g.attrs["data_comment"] = data_comment

    ###
    # "simulation_parameters" group
    ###
    g = f.create_group("simulation_parameters")
    g.attrs["refine_by"] = pf.refine_by
    g.attrs["dimensionality"] = pf.dimensionality
    g.attrs["domain_dimensions"] = pf.domain_dimensions
    g.attrs["current_time"] = pf.current_time
    g.attrs["domain_left_edge"] = pf.domain_left_edge
    g.attrs["domain_right_edge"] = pf.domain_right_edge
    g.attrs["unique_identifier"] = pf.unique_identifier
    g.attrs["cosmological_simulation"] = pf.cosmological_simulation
    # @todo: Where is this in the yt API?
    g.attrs["num_ghost_zones"] = 0
    # @todo: Where is this in the yt API?
    g.attrs["field_ordering"] = 0
    # @todo: not yet supported by yt.
    g.attrs["boundary_conditions"] = np.array([0, 0, 0, 0, 0, 0], 'int32')

    if pf.cosmological_simulation:
        g.attrs["current_redshift"] = pf.current_redshift
        g.attrs["omega_matter"] = pf.omega_matter
        g.attrs["omega_lambda"] = pf.omega_lambda
        g.attrs["hubble_constant"] = pf.hubble_constant

    ###
    # "field_types" group
    ###
    g = f.create_group("field_types")

    ###
    # "particle_types" group
    ###
    g = f.create_group("particle_types")

    # @todo: Particle type iterator
    sg = g.create_group(particle_type_name)
    sg["particle_type_name"] = particle_type_name

    ###
    # root datasets -- info about the grids
    ###
    f["grid_dimensions"] = pf.h.grid_dimensions
    f["grid_left_index"] = np.array(
            [g.get_global_startindex() for g in pf.h.grids]
    ).reshape(pf.h.grid_dimensions.shape[0], 3)
    f["grid_level"] = pf.h.grid_levels
    # @todo: Fill with proper values
    f["grid_parent_id"] = -np.ones(pf.h.grid_dimensions.shape[0])
    f["grid_particle_count"] = pf.h.grid_particle_count

    ###
    # "data" group -- where we should spend the most time
    ###
    
    g = f.create_group("data")
    for grid in pf.h.grids:
        # add group for this grid
        grid_group = g.create_group("grid_%010i" % (grid.id - grid._id_offset))
        # add group for the particles on this grid
        particles_group = grid_group.create_group("particles")
        pt_group = particles_group.create_group(particle_type_name)

    return f




"""
Author: John ZuHone <jzuhone@gmail.com>
Affiliation: NASA/GSFC
License:
  Copyright (C) 2012 John ZuHone All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.data_objects.data_containers import YTFieldData
from yt.data_objects.time_series import TimeSeriesData
from yt.utilities.lib import sample_field_at_positions
from yt.funcs import *

import numpy as np
import h5py

class ParticleTrajectoryCollection(object) :

    r"""A collection of particle trajectories in time over a series of
    parameter files. 

    The ParticleTrajectoryCollection object contains a collection of
    particle trajectories for a specified set of particle indices. 
    
    Parameters
    ----------
    filenames : list of strings
        A time-sorted list of filenames to construct the TimeSeriesData
        object.
    indices : array_like
        An integer array of particle indices whose trajectories we
        want to track. If they are not sorted they will be sorted.
    fields : list of strings, optional
        A set of fields that is retrieved when the trajectory
        collection is instantiated.
        Default : None (will default to the fields 'particle_position_x',
        'particle_position_y', 'particle_position_z')

    Examples
    ________
    >>> from yt.mods import *
    >>> my_fns = glob.glob("orbit_hdf5_chk_00[0-9][0-9]")
    >>> my_fns.sort()
    >>> fields = ["particle_position_x", "particle_position_y",
    >>>           "particle_position_z", "particle_velocity_x",
    >>>           "particle_velocity_y", "particle_velocity_z"]
    >>> pf = load(my_fns[0])
    >>> init_sphere = pf.h.sphere(pf.domain_center, (.5, "unitary"))
    >>> indices = init_sphere["particle_index"].astype("int")
    >>> trajs = ParticleTrajectoryCollection(my_fns, indices, fields=fields)
    >>> for t in trajs :
    >>>     print t["particle_velocity_x"].max(), t["particle_velocity_x"].min()

    Notes
    -----
    As of this time only particle trajectories that are complete over the
    set of specified parameter files are supported. If any particle's history
    ends for some reason (e.g. leaving the simulation domain or being actively
    destroyed), the whole trajectory collection of which it is a set must end
    at or before the particle's last timestep. This is a limitation we hope to
    lift at some point in the future.     
    """
    def __init__(self, filenames, indices, fields = None) :

        indices.sort() # Just in case the caller wasn't careful
        
        self.field_data = YTFieldData()
        self.pfs = TimeSeriesData.from_filenames(filenames)
        self.masks = []
        self.sorts = []
        self.indices = indices
        self.num_indices = len(indices)
        self.num_steps = len(filenames)
        self.times = []

        # Default fields 
        
        if fields is None : fields = []

        # Must ALWAYS have these fields
        
        fields = fields + ["particle_position_x",
                           "particle_position_y",
                           "particle_position_z"]

        """
        The following loops through the parameter files
        and performs two tasks. The first is to isolate
        the particles with the correct indices, and the
        second is to create a sorted list of these particles.
        We also make a list of the current time from each file. 
        Right now, the code assumes (and checks for) the
        particle indices existing in each file, a limitation I
        would like to lift at some point since some codes
        (e.g., FLASH) destroy particles leaving the domain.
        """
        
        for pf in self.pfs :
            dd = pf.h.all_data()
            newtags = dd["particle_index"].astype("int")
            if not np.all(np.in1d(indices, newtags, assume_unique=True)) :
                print "Not all requested particle ids contained in this file!"
                raise IndexError
            mask = np.in1d(newtags, indices, assume_unique=True)
            sorts = np.argsort(newtags[mask])
            self.masks.append(mask)            
            self.sorts.append(sorts)
            self.times.append(pf.current_time)

        self.times = np.array(self.times)

        # Set up the derived field list and the particle field list
        # so that if the requested field is a particle field, we'll
        # just copy the field over, but if the field is a grid field,
        # we will first copy the field over to the particle positions
        # and then return the field. 

        self.derived_field_list = self.pfs[0].h.derived_field_list
        self.particle_fields = [field for field in self.derived_field_list
                                if self.pfs[0].field_info[field].particle_type]

        # Now instantiate the requested fields 
        for field in fields :

            self._get_data(field)
            
    def has_key(self, key) :

        return (key in self.field_data)
    
    def keys(self) :

        return self.field_data.keys()

    def __getitem__(self, key) :
        """
        Get the field associated with key,
        checking to make sure it is a particle field.
        """

        if not self.field_data.has_key(key) :

            self._get_data(key)

        return self.field_data[key]
    
    def __setitem__(self, key, val):
        """
        Sets a field to be some other value.
        """
        self.field_data[key] = val
                        
    def __delitem__(self, key) :
        """
        Delete the field from the trajectory
        """
        del self.field_data[key]

    def __iter__(self) :

        """
        This iterates over the trajectories for
        the different particles, returning dicts
        of fields for each trajectory
        """
        for idx in xrange(self.num_indices) :
            traj = {}
            traj["particle_index"] = self.indices[idx]
            traj["particle_time"] = self.times
            for field in self.field_data.keys() :
                traj[field] = self[field][idx,:]
            yield traj
            
    def __len__(self) :

        """
        The number of individual trajectories
        """
        return self.num_indices

    def add_fields(self, fields) :

        """
        Add a list of fields to an existing trajectory

        Parameters
        ----------
        fields : list of strings
            A list of fields to be added to the current trajectory
            collection.

        Examples
        ________
        >>> from yt.mods import *
        >>> trajs = ParticleTrajectoryCollection(my_fns, indices)
        >>> trajs.add_fields(["particle_mass", "particle_gpot"])
        """
        
        for field in fields :

            if not self.field_data.has_key(field):

                self._get_data(field)
                
    def _get_data(self, field) :

        """
        Get a field to include in the trajectory collection.
        The trajectory collection itself is a dict of 2D numpy arrays,
        with shape (num_indices, num_steps)
        """
        
        if not self.field_data.has_key(field):
            
            particles = np.empty((0))

            step = int(0)
                
            for pf, mask, sort in zip(self.pfs, self.masks, self.sorts) :
                                    
                if field in self.particle_fields :

                    # This is easy... just get the particle fields

                    dd = pf.h.all_data()
                    pfield = dd[field][mask]
                    particles = np.append(particles, pfield[sort])

                else :

                    # This is hard... must loop over grids

                    pfield = np.zeros((self.num_indices))
                    x = self["particle_position_x"][:,step]
                    y = self["particle_position_y"][:,step]
                    z = self["particle_position_z"][:,step]

                    leaf_grids = [g for g in pf.h.grids if len(g.Children) == 0]
                        
                    for grid in leaf_grids :

                        pfield += sample_field_at_positions(grid[field],
                                                            grid.LeftEdge,
                                                            grid.RightEdge,
                                                            x, y, z)

                    particles = np.append(particles, pfield)

                step += 1
                
            self[field] = particles.reshape(self.num_steps,
                                            self.num_indices).transpose()

        return self.field_data[field]

    def trajectory_from_index(self, index) :

        """
        Retrieve a single trajectory corresponding to a specific particle
        index

        Parameters
        ----------
        index : int
            This defines which particle trajectory from the
            ParticleTrajectoryCollection object will be returned.

        Returns
        -------
        A dictionary corresponding to the particle's trajectory and the
        fields along that trajectory

        Examples
        --------
        >>> from yt.mods import *
        >>> import matplotlib.pylab as pl
        >>> trajs = ParticleTrajectoryCollection(my_fns, indices)
        >>> traj = trajs.trajectory_from_index(indices[0])
        >>> pl.plot(traj["particle_time"], traj["particle_position_x"], "-x")
        >>> pl.savefig("orbit")
        """
        
        mask = np.in1d(self.indices, (index,), assume_unique=True)

        if not np.any(mask) :
            print "The particle index %d is not in the list!" % (index)
            raise IndexError

        fields = [field for field in sorted(self.field_data.keys())]
                                
        traj = {}

        traj["particle_time"] = self.times
        traj["particle_index"] = index
        
        for field in fields :

            traj[field] = self[field][mask,:][0]

        return traj

    def write_out(self, filename_base) :

        """
        Write out particle trajectories to tab-separated ASCII files (one
        for each trajectory) with the field names in the file header. Each
        file is named with a basename and the index number.

        Parameters
        ----------
        filename_base : string
            The prefix for the outputted ASCII files.

        Examples
        --------
        >>> from yt.mods import *
        >>> trajs = ParticleTrajectoryCollection(my_fns, indices)
        >>> trajs.write_out("orbit_trajectory")       
        """
        
        fields = [field for field in sorted(self.field_data.keys())]

        num_fields = len(fields)

        first_str = "# particle_time\t" + "\t".join(fields)+"\n"
        
        template_str = "%g\t"*num_fields+"%g\n"
        
        for ix in xrange(self.num_indices) :

            outlines = [first_str]

            for it in xrange(self.num_steps) :
                outlines.append(template_str %
                                tuple([self.times[it]]+[self[field][ix,it] for field in fields]))
            
            fid = open(filename_base + "_%d.dat" % self.indices[ix], "w")
            fid.writelines(outlines)
            fid.close()
            del fid
            
    def write_out_h5(self, filename) :

        """
        Write out all the particle trajectories to a single HDF5 file
        that contains the indices, the times, and the 2D array for each
        field individually

        Parameters
        ----------

        filename : string
            The output filename for the HDF5 file

        Examples
        --------

        >>> from yt.mods import *
        >>> trajs = ParticleTrajectoryCollection(my_fns, indices)
        >>> trajs.write_out_h5("orbit_trajectories")                
        """
        
        fid = h5py.File(filename, "w")

        fields = [field for field in sorted(self.field_data.keys())]
        
        fid.create_dataset("particle_indices", dtype=np.int32,
                           data=self.indices)
        fid.create_dataset("particle_time", data=self.times)
        
        for field in fields :

            fid.create_dataset("%s" % field, data=self[field])
                        
        fid.close()



"""
Definitions specific to Nyx

Author: Casey W. Stark <caseywstark@gmail.com>
Affiliation: UC Berkeley
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Casey W. Stark, J. S. Oishi, Matthew Turk.  All Rights
  Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from utils import boxlib_bool_to_int

# This gives the type of each parameter we want, used to cast with a ``map()``
# @todo: get rid of enzo parameters we do not need
parameter_type_dict = {
    "CosmologyCurrentRedshift": float,
    "CosmologyComovingBoxSize": float,
    "CosmologyOmegaMatterNow": float,
    "CosmologyOmegaLambdaNow": float,
    "CosmologyHubbleConstantNow": float,
    "CosmologyInitialRedshift": float,
    "DualEnergyFormalismEta1": float,
    "DualEnergyFormalismEta2": float,
    "MetaDataString": str,
    "HydroMethod": int,
    "DualEnergyFormalism": int,
    "InitialTime": float,
    "ComovingCoordinates": boxlib_bool_to_int,
    "DensityUnits": float,
    "LengthUnits": float,
    "LengthUnit": float,
    "TemperatureUnits": float,
    "TimeUnits": float,
    "GravitationalConstant": float,
    "Gamma": float,
    "MultiSpecies": int,
    "CompilerPrecision": str,
    "CurrentTimeIdentifier": int,
    "RefineBy": int,
    "BoundaryConditionName": str,
    "TopGridRank": int,
    "TopGridDimensions": int,
    "EOSSoundSpeed": float,
    "EOSType": int,
    "NumberOfParticleAttributes": int,
}

# Provides translation between parameters in the nyx `inputs` file names to the
# enzo/yt name expected throughout the code. The key is nyx name, value is
# enzo/yt equivalent.
nyx_to_enzo_dict = {
    "amr.n_cell": "TopGridDimensions",
    "amr.ref_ratio": "RefineBy",
    "materials.gamma": "Gamma",
    "castro.use_comoving": "ComovingCoordinates",
    "castro.redshift_in": "CosmologyInitialRedshift",
    "comoving_OmL": "CosmologyOmegaLambdaNow",
    "comoving_OmM": "CosmologyOmegaMatterNow",
    "comoving_h": "CosmologyHubbleConstantNow"
}

# is this the same as nyx_to_enzo.
yt_to_nyx_fields_dict = {}
nyx_to_yt_fields_dict = {}

fab_header_pattern = r"^FAB \(\((\d+), \([0-9 ]+\)\),\(\d+, \(([0-9 ]+)\)\)\)\(\((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\)\) (\d+)\n"

# need to specify units eventually
nyx_particle_field_names = ['particle_position_%s' % ax for ax in 'xyz'] + \
                           ['particle_mass'] +  \
                           ['particle_velocity_%s' % ax for ax in 'xyz']




"""
Base classes for answer testing

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import matplotlib
from yt.mods import *

# We first create our dictionary of tests to run.  This starts out empty, and
# as tests are imported it will be filled.
if "TestRegistry" not in locals():
    class TestRegistry(dict):
        def __new__(cls, *p, **k):
            if not '_the_instance' in cls.__dict__:
                cls._the_instance = dict.__new__(cls)
                return cls._the_instance
if "test_registry" not in locals():
    test_registry = TestRegistry()

# The exceptions we raise, related to the character of the failure.

class RegressionTestException(Exception):
    pass

class ValueDelta(RegressionTestException):
    def __init__(self, delta, acceptable):
        self.delta = delta
        self.acceptable = acceptable

    def __repr__(self):
        return "ValueDelta: Delta %s, max of %s" % (
            self.delta, self.acceptable)

class ArrayDelta(ValueDelta):
    def __repr__(self):
        nabove = len(np.where(self.delta > self.acceptable)[0])
        return "ArrayDelta: Delta max of %s, acceptable of %s.\n" \
               "%d of %d points above the acceptable limit" % \
               (np.nanmax(self.delta), self.acceptable, nabove,
                self.delta.size)

class ShapeMismatch(RegressionTestException):
    def __init__(self, old_shape, current_shape):
        self.old_shape = old_shape
        self.current_shape = current_shape

    def __repr__(self):
        return "Shape Mismatch: old_buffer %s, current_buffer %s" % (
            self.old_shape, self.current_shape)

class RegressionTest(object):
    name = None
    result = None
    output_type = None

    class __metaclass__(type):
        # This ensures that all the tests are auto-registered if they have a
        # name.  If they do not have a name, they are considered to be base
        # classes to be overridden and implemented by someone else.
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if cls.name is not None:
                test_registry[cls.name] = cls

    def setup(self):
        """
        This function must be defined if the problem requires additional setup.
        Note that for the most part this will be defined in base classes where
        subclasses will only implement 'run'.
        """
        pass

    def run(self):
        """
        This function must generate a result value, of any type, and store it
        in self.result.
        """
        pass

    def compare(self, old_result):
        """
        This function must accept `old_result` and compare it somehow against
        the value stored in `self.result`.  If the result is a failure, it must
        raise an exception.  Otherwise it is considered to be a success.
        """
        pass

    def plot(self):
        """
        This function can optionally plot the contents of `self.result`.
        """
        pass

    def compare_array_delta(self, a1, a2, acceptable):
        """
        This is a helper function.  It accepts two numpy arrays and compares
        the maximum relative difference.  If the maximum relative difference is
        greater than `acceptable` it is considered a failure and an appropriate
        exception is raised.
        """
        if a1.shape != a2.shape:
            raise ShapeMismatch(a1, a2)
        delta = np.abs(a1 - a2).astype("float64")/(a1 + a2)
        if np.nanmax(delta) > acceptable:
            raise ArrayDelta(delta, acceptable)
        return True

    def compare_value_delta(self, v1, v2, acceptable):
        """
        This is a helper function.  It accepts two floating point values and
        calculates the maximum relative different.  If the maximum relative
        difference is greater than `acceptable` it is considered a failure and
        an appropriate exception is raised.
        """
        delta = np.abs(v1 - v2)/(v1 + v2)
        if delta > acceptable:
            raise ValueDelta(delta, acceptable)
        return True

class SingleOutputTest(RegressionTest):
    output_type = 'single'

    def __init__(self, filename):
        """
        This test mechanism is designed to accept a single filename and
        evaluate it, not necessarily utilizing yt's functionality to do so.
        """
        self.filename = filename

class MultipleOutputTest(RegressionTest):
    output_type = 'multiple'

    io_log_header = "DATASET WRITTEN"

    def __init__(self, io_log):
        """
        This test mechanism is designed to accept an OutputLog file and then
        iterate over it, evaluating every single dataset individually.
        """
        self.io_log = io_log

    def __iter__(self):
        if isinstance(self.io_log, types.StringTypes):
            for line in open(self.io_log):
                yield line[len(self.io_log_header):].split()[0].strip()
        elif isinstance(self.io_log, types.ListType):
            for line in self.io_log: yield line

def create_test(base, new_name, **attrs):
    """
    This function accepts a base class of a test, sets some attributes on it,
    and then registers a new test.  It's a fast way of registering multiple
    tests that share the same testing logic but that differ on a few parameters
    or combinations of parameters.
    """
    new_name = "%s_%s" % (base.__name__, new_name)
    attrs['name'] = new_name
    return type(new_name, (base,), attrs)

class YTStaticOutputTest(SingleOutputTest):

    def setup(self):
        self.pf = load(self.filename)

    def pixelize(self, data, field, edges = None, dims = (512, 512)):
        """
        This is a helper function that returns a 2D array of the specified
        source, in the specified field, at the specified spatial extent.
        """
        xax = x_dict[self.axis]
        yax = y_dict[self.axis]
        
        if edges is None:
            edges = (self.pf.domain_left_edge[xax],
                     self.pf.domain_right_edge[xax],
                     self.pf.domain_left_edge[yax],
                     self.pf.domain_right_edge[yax])
        frb = FixedResolutionBuffer( data, edges, dims)
        frb[field] # To make the pixelization
        return frb

    def compare_data_arrays(self, d1, d2, tol = 1e-7):
        """
        This is a helper function.  It accepts two dictionaries of numpy arrays
        and compares the maximum relative difference of every array.  If the
        maximum relative difference is greater than `acceptable` it is
        considered a failure and an appropriate exception is raised.
        """
        for field in d1.keys():
            self.compare_array_delta(d1[field], d2[field], tol)

    @property
    def sim_center(self):
        """
        This returns the center of the domain.
        """
        return 0.5*(self.pf.domain_right_edge + self.pf.domain_left_edge)

    @property
    def max_dens_location(self):
        """
        This is a helper function to return the location of the most dense
        point.
        """
        return self.pf.h.find_max("Density")[1]

    @property
    def entire_simulation(self):
        """
        Return an unsorted array of values that cover the entire domain.
        """
        return self.pf.h.all_data()
        




"""
Fields specific to Enzo

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    NullFunc, \
    TranslationFunc, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields
from yt.utilities.physical_constants import \
    mh, \
    kpc_per_cm
from yt.funcs import *

import yt.utilities.lib as amr_utils

EnzoFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = EnzoFieldInfo.add_field

KnownEnzoFields = FieldInfoContainer()
add_enzo_field = KnownEnzoFields.add_field

_speciesList = ["HI", "HII", "Electron",
                "HeI", "HeII", "HeIII",
                "H2I", "H2II", "HM",
                "DI", "DII", "HDI", "Metal", "MetalSNIa", "PreShock"]
_speciesMass = {"HI": 1.0, "HII": 1.0, "Electron": 1.0,
                "HeI": 4.0, "HeII": 4.0, "HeIII": 4.0,
                "H2I": 2.0, "H2II": 2.0, "HM": 1.0,
                "DI": 2.0, "DII": 2.0, "HDI": 3.0}

def _SpeciesComovingDensity(field, data):
    sp = field.name.split("_")[0] + "_Density"
    ef = (1.0 + data.pf.current_redshift)**3.0
    return data[sp] / ef

def _SpeciesFraction(field, data):
    sp = field.name.split("_")[0] + "_Density"
    return data[sp] / data["Density"]

def _SpeciesMass(field, data):
    sp = field.name.split("_")[0] + "_Density"
    return data[sp] * data["CellVolume"]

def _SpeciesNumberDensity(field, data):
    species = field.name.split("_")[0]
    sp = field.name.split("_")[0] + "_Density"
    return data[sp] / _speciesMass[species]

def _convertCellMassMsun(data):
    return 5.027854e-34 # g^-1

def _ConvertNumberDensity(data):
    return 1.0/mh

for species in _speciesList:
    add_field("%s_Fraction" % species,
             function=_SpeciesFraction,
             validators=ValidateDataField("%s_Density" % species),
             display_name="%s\/Fraction" % species)
    add_field("Comoving_%s_Density" % species,
             function=_SpeciesComovingDensity,
             validators=ValidateDataField("%s_Density" % species),
             display_name="Comoving\/%s\/Density" % species)
    add_field("%s_Mass" % species, units=r"\rm{g}", 
              function=_SpeciesMass, 
              validators=ValidateDataField("%s_Density" % species),
              display_name="%s\/Mass" % species)
    add_field("%s_MassMsun" % species, units=r"M_{\odot}", 
              function=_SpeciesMass, 
              convert_function=_convertCellMassMsun,
              validators=ValidateDataField("%s_Density" % species),
              display_name="%s\/Mass" % species)
    if _speciesMass.has_key(species):
        add_field("%s_NumberDensity" % species,
                  function=_SpeciesNumberDensity,
                  convert_function=_ConvertNumberDensity,
                  validators=ValidateDataField("%s_Density" % species))

def _Metallicity(field, data):
    return data["Metal_Fraction"]
def _ConvertMetallicity(data):
    return 49.0196 # 1 / 0.0204
add_field("Metallicity", units=r"Z_{\rm{\odot}}",
          function=_Metallicity,
          convert_function=_ConvertMetallicity,
          validators=ValidateDataField("Metal_Density"),
          projection_conversion="1")

def _Metallicity3(field, data):
    return data["SN_Colour"]/data["Density"]
add_field("Metallicity3", units=r"Z_{\rm{\odot}}",
          function=_Metallicity3,
          convert_function=_ConvertMetallicity,
          validators=ValidateDataField("SN_Colour"),
          projection_conversion="1")

add_enzo_field("Cooling_Time", units=r"\rm{s}",
               function=NullFunc,
               validators=ValidateDataField("Cooling_Time"),
               projection_conversion="1")

def _ThermalEnergy(field, data):
    if data.pf["HydroMethod"] == 2:
        return data["TotalEnergy"]
    
    if data.pf["DualEnergyFormalism"]:
        return data["GasEnergy"]

    if data.pf["HydroMethod"] in (4,6):
        return data["TotalEnergy"] - 0.5*(
            data["x-velocity"]**2.0
            + data["y-velocity"]**2.0
            + data["z-velocity"]**2.0 ) \
            - data["MagneticEnergy"]/data["Density"]

    return data["TotalEnergy"] - 0.5*(
        data["x-velocity"]**2.0
        + data["y-velocity"]**2.0
        + data["z-velocity"]**2.0 )
add_field("ThermalEnergy", function=_ThermalEnergy,
          units=r"\rm{ergs}/\rm{g}")

def _KineticEnergy(field, data):
    return 0.5*data["Density"] * ( data["x-velocity"]**2.0
                                   + data["y-velocity"]**2.0
                                   + data["z-velocity"]**2.0 )
add_field("KineticEnergy",function=_KineticEnergy,
          units = r"\rm{ergs}/\rm{cm^3}")
# This next section is the energy field section
# Note that we have aliases that manually unconvert themselves.
# This is because numerous code branches use Gas_Energy or GasEnergy
# indiscriminately -- this is almost fixed with LCA1.5, but not everyone is
# moving to that branch.  So, because the actual function doesn't get called
# *unless* it's an alias, we simply de-convert -- since the input data is
# already converted to cgs.

def _convertEnergy(data):
    return data.convert("x-velocity")**2.0

add_enzo_field("GasEnergy", function=NullFunc,
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)
add_enzo_field("Gas_Energy", function=NullFunc,
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _Gas_Energy(field, data):
    return data["GasEnergy"] / _convertEnergy(data)
add_field("Gas_Energy", function=_Gas_Energy,
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

# We set up fields for both TotalEnergy and Total_Energy in the known fields
# lists.  Note that this does not mean these will be the used definitions.
add_enzo_field("TotalEnergy", function=NullFunc,
          display_name = r"\rm{Total}\/ \rm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)
add_enzo_field("Total_Energy", function=NullFunc,
          display_name = r"\rm{Total}\/ \rm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _Total_Energy(field, data):
    return data["TotalEnergy"] / _convertEnergy(data)
add_field("Total_Energy", function=_Total_Energy,
          display_name = r"\rm{Total}\/ \rm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _TotalEnergy(field, data):
    return data["Total_Energy"] / _convertEnergy(data)
add_field("TotalEnergy", function=_TotalEnergy,
          display_name = r"\rm{Total}\/ \rm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _NumberDensity(field, data):
    # We can assume that we at least have Density
    # We should actually be guaranteeing the presence of a .shape attribute,
    # but I am not currently implementing that
    fieldData = np.zeros(data["Density"].shape,
                         dtype = data["Density"].dtype)
    if data.pf["MultiSpecies"] == 0:
        if data.has_field_parameter("mu"):
            mu = data.get_field_parameter("mu")
        else:
            mu = 0.6
        fieldData += data["Density"] / mu
    if data.pf["MultiSpecies"] > 0:
        fieldData += data["HI_Density"] / 1.0
        fieldData += data["HII_Density"] / 1.0
        fieldData += data["HeI_Density"] / 4.0
        fieldData += data["HeII_Density"] / 4.0
        fieldData += data["HeIII_Density"] / 4.0
        fieldData += data["Electron_Density"] / 1.0
    if data.pf["MultiSpecies"] > 1:
        fieldData += data["HM_Density"] / 1.0
        fieldData += data["H2I_Density"] / 2.0
        fieldData += data["H2II_Density"] / 2.0
    if data.pf["MultiSpecies"] > 2:
        fieldData += data["DI_Density"] / 2.0
        fieldData += data["DII_Density"] / 2.0
        fieldData += data["HDI_Density"] / 3.0
    return fieldData
add_field("NumberDensity", units=r"\rm{cm}^{-3}",
          function=_NumberDensity,
          convert_function=_ConvertNumberDensity)

def _H_NumberDensity(field, data):
    field_data = np.zeros(data["Density"].shape,
                          dtype=data["Density"].dtype)
    if data.pf.parameters["MultiSpecies"] == 0:
        field_data += data["Density"] * \
          data.pf.parameters["HydrogenFractionByMass"]
    if data.pf.parameters["MultiSpecies"] > 0:
        field_data += data["HI_Density"]
        field_data += data["HII_Density"]
    if data.pf.parameters["MultiSpecies"] > 1:
        field_data += data["HM_Density"]
        field_data += data["H2I_Density"]
        field_data += data["H2II_Density"]
    if data.pf.parameters["MultiSpecies"] > 2:
        field_data += data["HDI_Density"] / 2.0
    return field_data
add_field("H_NumberDensity", units=r"\rm{cm}^{-3}",
          function=_H_NumberDensity,
          convert_function=_ConvertNumberDensity)


# Now we add all the fields that we want to control, but we give a null function
# This is every Enzo field we can think of.  This will be installation-dependent,

# removed: "Gas_Energy","Total_Energy",
# these are now aliases for each other

_default_fields = ["Density","Temperature",
                   "x-velocity","y-velocity","z-velocity",
                   "x-momentum","y-momentum","z-momentum",
                   "Bx", "By", "Bz", "Dust_Temperature",
                   "HI_kph", "HeI_kph", "HeII_kph", "H2I_kdiss", "PhotoGamma",
                   "RadAccel1", "RadAccel2", "RadAccel3", "SN_Colour",
                   "Ray_Segments"]
# else:
#     _default_fields = ["Density","Temperature","Gas_Energy","Total_Energy",
#                        "x-velocity","y-velocity","z-velocity"]
_default_fields += [ "%s_Density" % sp for sp in _speciesList ]

for field in _default_fields:
    dn = field.replace("_","\/")
    add_enzo_field(field, function=NullFunc, take_log=True,
              display_name = dn,
              validators=[ValidateDataField(field)], units=r"Unknown")
KnownEnzoFields["x-velocity"].projection_conversion='1'
KnownEnzoFields["y-velocity"].projection_conversion='1'
KnownEnzoFields["z-velocity"].projection_conversion='1'

def _convertBfield(data): 
    return np.sqrt(4*np.pi*data.convert("Density")*data.convert("x-velocity")**2)
for field in ['Bx','By','Bz']:
    f = KnownEnzoFields[field]
    f._convert_function=_convertBfield
    f._units=r"\rm{Gauss}"
    f.take_log=False

def _convertRadiation(data):
    return 1.0/data.convert("Time")
for field in ["HI_kph", "HeI_kph", "HeII_kph", "H2I_kdiss"]:
    f = KnownEnzoFields[field]
    f._convert_function = _convertRadiation
    f._units=r"\rm{s}^{-1}"
    f.take_log=True

KnownEnzoFields["PhotoGamma"]._convert_function = _convertRadiation
KnownEnzoFields["PhotoGamma"]._units = r"\rm{eV} \rm{s}^{-1}"
KnownEnzoFields["PhotoGamma"].take_log = True

def _convertRadiationAccel(data):
    return data.convert("cm") / data.convert("Time")**2
for dim in range(1,4):
    f = KnownEnzoFields["RadAccel%d" % dim]
    f._convert_function = _convertRadiationAccel
    f._units=r"\rm{cm}\/\rm{s}^{-2}"
    f.take_log=False
def _RadiationAccelerationMagnitude(field, data):
    return ( data["RadAccel1"]**2 + data["RadAccel2"]**2 +
             data["RadAccel3"]**2 )**(1.0/2.0)
add_field("RadiationAcceleration", 
          function=_RadiationAccelerationMagnitude,
          validators=ValidateDataField(["RadAccel1", "RadAccel2", "RadAccel3"]),
          display_name="Radiation\/Acceleration", units=r"\rm{cm} \rm{s}^{-2}")

# Now we override

def _convertDensity(data):
    return data.convert("Density")

def _convertCmToKpc(data):
    return 1/(kpc_per_cm)**3

for field in ["Density"] + [ "%s_Density" % sp for sp in _speciesList ] + \
        ["SN_Colour"]:
    KnownEnzoFields[field]._units = r"\rm{g}/\rm{cm}^3"
    KnownEnzoFields[field]._projected_units = r"\rm{g}/\rm{cm}^2"
    KnownEnzoFields[field]._convert_function=_convertDensity

add_enzo_field("Dark_Matter_Density", function=NullFunc,
          convert_function=_convertDensity,
          validators=[ValidateDataField("Dark_Matter_Density"),
                      ValidateSpatial(0)],
          display_name = "Dark\/Matter\/Density",
          not_in_all = True)

def _Dark_Matter_Mass(field, data):
    return data['Dark_Matter_Density'] * data["CellVolume"]
add_field("Dark_Matter_Mass", function=_Dark_Matter_Mass,
          validators=ValidateDataField("Dark_Matter_Density"),
          display_name="Dark\/Matter\/Mass", units=r"\rm{g}")
add_field("Dark_Matter_MassMsun", function=_Dark_Matter_Mass,
          convert_function=_convertCellMassMsun,
          validators=ValidateDataField("Dark_Matter_Density"),
          display_name="Dark\/Matter\/Mass", units=r"M_{\odot}")

KnownEnzoFields["Temperature"]._units = r"\rm{K}"
KnownEnzoFields["Temperature"].units = r"K"
KnownEnzoFields["Dust_Temperature"]._units = r"\rm{K}"
KnownEnzoFields["Dust_Temperature"].units = r"K"

def _convertVelocity(data):
    return data.convert("x-velocity")
for ax in ['x','y','z']:
    f = KnownEnzoFields["%s-velocity" % ax]
    f._units = r"\rm{cm}/\rm{s}"
    f._convert_function = _convertVelocity
    f.take_log = False

def _spdensity(field, data):
    blank = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return blank
    filter = data['creation_time'] > 0.0
    if not filter.any(): return blank
    amr_utils.CICDeposit_3(data["particle_position_x"][filter].astype(np.float64),
                           data["particle_position_y"][filter].astype(np.float64),
                           data["particle_position_z"][filter].astype(np.float64),
                           data["particle_mass"][filter],
                           np.int64(np.where(filter)[0].size),
                           blank, np.array(data.LeftEdge).astype(np.float64),
                           np.array(data.ActiveDimensions).astype(np.int32), 
                           np.float64(data['dx']))
    return blank

add_field("star_density", function=_spdensity,
          validators=[ValidateSpatial(0)], convert_function=_convertDensity,
          units = r"\rm{g}/\rm{cm}^3",
          projected_units = r"\rm{g}/\rm{cm}^2",
          display_name = "Stellar\/Density")

def _tpdensity(field, data): 
    blank = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return blank
    filter = data['particle_type'] == 3 # tracer particles
    if not filter.any(): return blank
    amr_utils.CICDeposit_3(data["particle_position_x"][filter].astype(np.float64),
                           data["particle_position_y"][filter].astype(np.float64),
                           data["particle_position_z"][filter].astype(np.float64),
                           np.ones(filter.sum(), dtype="float64"),
                           np.int64(np.where(filter)[0].size),
                           blank, np.array(data.LeftEdge).astype(np.float64),
                           np.array(data.ActiveDimensions).astype(np.int32), 
                           np.float64(data['dx']))
    blank /= data['CellVolume']
    return blank

add_field("tracer_number_density", function=_tpdensity,
          validators=[ValidateSpatial(0)], convert_function=_convertCmToKpc,
          units = r"\rm{particles}/\rm{kpc}^3",
          projected_units = r"\rm{particles}/\rm{kpc}^2",
          display_name = "Tracer\/Particle\/Number\/Density",
          projection_conversion='kpc')

def _dmpdensity(field, data):
    blank = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return blank
    if 'creation_time' in data.pf.field_info:
        filter = data['creation_time'] <= 0.0
        if not filter.any(): return blank
        num = filter.sum()
    else:
        filter = Ellipsis
        num = data["particle_position_x"].size
    amr_utils.CICDeposit_3(data["particle_position_x"][filter].astype(np.float64),
                           data["particle_position_y"][filter].astype(np.float64),
                           data["particle_position_z"][filter].astype(np.float64),
                           data["particle_mass"][filter],
                           num,
                           blank, np.array(data.LeftEdge).astype(np.float64),
                           np.array(data.ActiveDimensions).astype(np.int32), 
                           np.float64(data['dx']))
    return blank

add_field("dm_density", function=_dmpdensity,
          validators=[ValidateSpatial(0)], convert_function=_convertDensity,
          units = r"\rm{g}/\rm{cm}^3",
          projected_units = r"\rm{g}/\rm{cm}^2",
          display_name = "Dark\/Matter\/Density")

def _cic_particle_field(field, data):
    """
    Create a grid field for particle quantities weighted by particle mass, 
    using cloud-in-cell deposit.
    """
    particle_field = field.name[4:]
    top = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return top
    particle_field_data = data[particle_field] * data['particle_mass']
    amr_utils.CICDeposit_3(data["particle_position_x"].astype(np.float64),
                           data["particle_position_y"].astype(np.float64),
                           data["particle_position_z"].astype(np.float64),
                           particle_field_data,
                           data["particle_position_x"].size,
                           top, np.array(data.LeftEdge).astype(np.float64),
                           np.array(data.ActiveDimensions).astype(np.int32), 
                           np.float64(data['dx']))
    del particle_field_data

    bottom = np.zeros(data.ActiveDimensions, dtype='float64')
    amr_utils.CICDeposit_3(data["particle_position_x"].astype(np.float64),
                           data["particle_position_y"].astype(np.float64),
                           data["particle_position_z"].astype(np.float64),
                           data["particle_mass"],
                           data["particle_position_x"].size,
                           bottom, np.array(data.LeftEdge).astype(np.float64),
                           np.array(data.ActiveDimensions).astype(np.int32), 
                           np.float64(data['dx']))
    top[bottom == 0] = 0.0
    bnz = bottom.nonzero()
    top[bnz] /= bottom[bnz]
    return top

add_field('cic_particle_velocity_x', function=_cic_particle_field,
          take_log=False, validators=[ValidateSpatial(0)])
add_field('cic_particle_velocity_y', function=_cic_particle_field,
          take_log=False, validators=[ValidateSpatial(0)])
add_field('cic_particle_velocity_z', function=_cic_particle_field,
          take_log=False, validators=[ValidateSpatial(0)])

def _star_field(field, data):
    """
    Create a grid field for star quantities, weighted by star mass.
    """
    particle_field = field.name[5:]
    top = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return top
    filter = data['creation_time'] > 0.0
    if not filter.any(): return top
    particle_field_data = data[particle_field][filter] * data['particle_mass'][filter]
    amr_utils.CICDeposit_3(data["particle_position_x"][filter].astype(np.float64),
                          data["particle_position_y"][filter].astype(np.float64),
                          data["particle_position_z"][filter].astype(np.float64),
                          particle_field_data,
                          np.int64(np.where(filter)[0].size),
                          top, np.array(data.LeftEdge).astype(np.float64),
                          np.array(data.ActiveDimensions).astype(np.int32), 
                          np.float64(data['dx']))
    del particle_field_data

    bottom = np.zeros(data.ActiveDimensions, dtype='float64')
    amr_utils.CICDeposit_3(data["particle_position_x"][filter].astype(np.float64),
                          data["particle_position_y"][filter].astype(np.float64),
                          data["particle_position_z"][filter].astype(np.float64),
                          data["particle_mass"][filter],
                          np.int64(np.where(filter)[0].size),
                          bottom, np.array(data.LeftEdge).astype(np.float64),
                          np.array(data.ActiveDimensions).astype(np.int32), 
                          np.float64(data['dx']))
    top[bottom == 0] = 0.0
    bnz = bottom.nonzero()
    top[bnz] /= bottom[bnz]
    return top

add_field('star_metallicity_fraction', function=_star_field,
          validators=[ValidateSpatial(0)])
add_field('star_creation_time', function=_star_field,
          validators=[ValidateSpatial(0)])
add_field('star_dynamical_time', function=_star_field,
          validators=[ValidateSpatial(0)])

def _StarMetallicity(field, data):
    return data['star_metallicity_fraction']
add_field('StarMetallicity', units=r"Z_{\rm{\odot}}",
          function=_StarMetallicity,
          convert_function=_ConvertMetallicity,
          projection_conversion="1")

def _StarCreationTime(field, data):
    return data['star_creation_time']
def _ConvertEnzoTimeYears(data):
    return data.pf.time_units['years']
add_field('StarCreationTimeYears', units=r"\rm{yr}",
          function=_StarCreationTime,
          convert_function=_ConvertEnzoTimeYears,
          projection_conversion="1")

def _StarDynamicalTime(field, data):
    return data['star_dynamical_time']
add_field('StarDynamicalTimeYears', units=r"\rm{yr}",
          function=_StarDynamicalTime,
          convert_function=_ConvertEnzoTimeYears,
          projection_conversion="1")

def _StarAge(field, data):
    star_age = np.zeros(data['StarCreationTimeYears'].shape)
    with_stars = data['StarCreationTimeYears'] > 0
    star_age[with_stars] = data.pf.time_units['years'] * \
        data.pf.current_time - \
        data['StarCreationTimeYears'][with_stars]
    return star_age
add_field('StarAgeYears', units=r"\rm{yr}",
          function=_StarAge,
          projection_conversion="1")

def _IsStarParticle(field, data):
    is_star = (data['creation_time'] > 0).astype('float64')
    return is_star
add_field('IsStarParticle', function=_IsStarParticle,
          particle_type = True)

def _Bmag(field, data):
    """ magnitude of bvec
    """
    return np.sqrt(data['Bx']**2 + data['By']**2 + data['Bz']**2)

add_field("Bmag", function=_Bmag,display_name=r"$|B|$",units=r"\rm{Gauss}")

# Particle functions

def particle_func(p_field, dtype='float64'):
    def _Particles(field, data):
        io = data.hierarchy.io
        if not data.NumberOfParticles > 0:
            return np.array([], dtype=dtype)
        try:
            return io._read_data_set(data, p_field).astype(dtype)
        except io._read_exception:
            pass
        # This is bad.  But it's the best idea I have right now.
        return data._read_data(p_field.replace("_"," ")).astype(dtype)
    return _Particles
for pf in ["type", "mass"] + \
          ["position_%s" % ax for ax in 'xyz']:
    pfunc = particle_func("particle_%s" % (pf))
    add_enzo_field("particle_%s" % pf, function=pfunc,
              validators = [ValidateSpatial(0)],
              particle_type=True)
    
def _convRetainInt(data):
    return 1
add_enzo_field("particle_index", function=particle_func("particle_index", "int64"),
          validators = [ValidateSpatial(0)], particle_type=True,
          convert_function=_convRetainInt)

def _get_vel_convert(ax):
    def _convert_p_vel(data):
        return data.convert("%s-velocity" % ax)
    return _convert_p_vel
for ax in 'xyz':
    pf = "particle_velocity_%s" % ax
    pfunc = particle_func(pf)
    cfunc = _get_vel_convert(ax)
    add_enzo_field(pf, function=pfunc, convert_function=cfunc,
              validators = [ValidateSpatial(0)],
              particle_type=True)

for pf in ["creation_time", "dynamical_time", "metallicity_fraction"]:
    pfunc = particle_func(pf)
    add_enzo_field(pf, function=pfunc,
              validators = [ValidateSpatial(0),
                            ValidateDataField(pf)],
              particle_type=True)
add_field("particle_mass", function=particle_func("particle_mass"),
          validators=[ValidateSpatial(0)], particle_type=True)

def _ParticleAge(field, data):
    current_time = data.pf.current_time
    return (current_time - data["creation_time"])
def _convertParticleAge(data):
    return data.convert("years")
add_field("ParticleAge", function=_ParticleAge,
          validators=[ValidateDataField("creation_time")],
          particle_type=True, convert_function=_convertParticleAge)

def _ParticleMass(field, data):
    particles = data["particle_mass"].astype('float64') * \
                just_one(data["CellVolumeCode"].ravel())
    # Note that we mandate grid-type here, so this is okay
    return particles

def _convertParticleMass(data):
    return data.convert("Density")*(data.convert("cm")**3.0)
def _IOLevelParticleMass(grid):
    dd = dict(particle_mass = np.ones(1), CellVolumeCode=grid["CellVolumeCode"])
    cf = (_ParticleMass(None, dd) * _convertParticleMass(grid))[0]
    return cf
def _convertParticleMassMsun(data):
    return data.convert("Density")*((data.convert("cm")**3.0)/1.989e33)
def _IOLevelParticleMassMsun(grid):
    dd = dict(particle_mass = np.ones(1), CellVolumeCode=grid["CellVolumeCode"])
    cf = (_ParticleMass(None, dd) * _convertParticleMassMsun(grid))[0]
    return cf
add_field("ParticleMass",
          function=_ParticleMass, validators=[ValidateSpatial(0)],
          particle_type=True, convert_function=_convertParticleMass,
          particle_convert_function=_IOLevelParticleMass)
add_field("ParticleMassMsun",
          function=_ParticleMass, validators=[ValidateSpatial(0)],
          particle_type=True, convert_function=_convertParticleMassMsun,
          particle_convert_function=_IOLevelParticleMassMsun)

#
# Now we do overrides for 2D fields
#

Enzo2DFieldInfo = FieldInfoContainer.create_with_fallback(EnzoFieldInfo)
add_enzo_2d_field = Enzo2DFieldInfo.add_field

def _CellArea(field, data):
    if data['dx'].size == 1:
        try:
            return data['dx']*data['dy']*\
                np.ones(data.ActiveDimensions, dtype='float64')
        except AttributeError:
            return data['dx']*data['dy']
    return data["dx"]*data["dy"]
def _ConvertCellAreaMpc(data):
    return data.convert("mpc")**2.0
def _ConvertCellAreaCGS(data):
    return data.convert("cm")**2.0
add_enzo_2d_field("CellAreaCode", units=r"\rm{BoxArea}^2",
          function=_CellArea)
add_enzo_2d_field("CellAreaMpc", units=r"\rm{Mpc}^2",
          function=_CellArea,
          convert_function=_ConvertCellAreaMpc)
add_enzo_2d_field("CellArea", units=r"\rm{cm}^2",
          function=_CellArea,
          convert_function=_ConvertCellAreaCGS)

for a in ["Code", "Mpc", ""]:
    Enzo2DFieldInfo["CellVolume%s" % a] = \
        Enzo2DFieldInfo["CellArea%s" % a]

def _zvel(field, data):
    return np.zeros(data["x-velocity"].shape,
                    dtype='float64')
add_enzo_2d_field("z-velocity", function=_zvel)

#
# Now we do overrides for 1D fields
#

Enzo1DFieldInfo = FieldInfoContainer.create_with_fallback(EnzoFieldInfo)
add_enzo_1d_field = Enzo1DFieldInfo.add_field

def _CellLength(field, data):
    return data["dx"]
def _ConvertCellLengthMpc(data):
    return data.convert("mpc")
def _ConvertCellLengthCGS(data):
    return data.convert("cm")
add_enzo_1d_field("CellLengthCode", units=r"\rm{BoxArea}^2",
          function=_CellLength)
add_enzo_1d_field("CellLengthMpc", units=r"\rm{Mpc}^2",
          function=_CellLength,
          convert_function=_ConvertCellLengthMpc)
add_enzo_1d_field("CellLength", units=r"\rm{cm}^2",
          function=_CellLength,
          convert_function=_ConvertCellLengthCGS)

for a in ["Code", "Mpc", ""]:
    Enzo1DFieldInfo["CellVolume%s" % a] = \
        Enzo1DFieldInfo["CellLength%s" % a]

def _yvel(field, data):
    return np.zeros(data["x-velocity"].shape,
                    dtype='float64')
add_enzo_1d_field("z-velocity", function=_zvel)
add_enzo_1d_field("y-velocity", function=_yvel)

